import asyncio
import json
import os
import posixpath
import shlex
from abc import ABC, abstractmethod
from typing import Any, MutableMapping, MutableSequence, Optional, Tuple

import pkg_resources
from cachetools import Cache, TTLCache

from streamflow.core import utils
from streamflow.core.asyncache import cachedmethod
from streamflow.core.context import StreamFlowContext
from streamflow.core.deployment import Connector
from streamflow.core.exception import WorkflowExecutionException
from streamflow.core.scheduling import Location
from streamflow.deployment.connector.base import BaseConnector
from streamflow.log_handler import logger


async def _exists_docker_image(image_name: str) -> bool:
    exists_command = "".join(
        "docker "
        "image "
        "inspect "
        "{image_name}"
    ).format(
        image_name=image_name
    )
    proc = await asyncio.create_subprocess_exec(
        *shlex.split(exists_command),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)
    await proc.wait()
    return proc.returncode == 0


async def _pull_docker_image(image_name: str) -> None:
    exists_command = "".join(
        "docker "
        "pull "
        "--quiet "
        "{image_name}"
    ).format(
        image_name=image_name
    )
    proc = await asyncio.create_subprocess_exec(
        *shlex.split(exists_command),
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.DEVNULL)
    await proc.wait()


class ContainerConnector(BaseConnector, ABC):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 transferBufferSize: int,
                 locationsCacheSize: int = None,
                 locationsCacheTTL: int = None,
                 resourcesCacheSize: int = None,
                 resourcesCacheTTL: int = None):
        super().__init__(deployment_name, context, transferBufferSize)
        cacheSize = locationsCacheSize
        if cacheSize is None:
            cacheSize = resourcesCacheSize
            if cacheSize is not None:
                logger.warn("The `resourcesCacheSize` keyword is deprecated and will be removed in StreamFlow 0.3.0. "
                            "Use `locationsCacheSize` instead.")
            else:
                cacheSize = 10
        cacheTTL = locationsCacheTTL
        if cacheTTL is None:
            cacheTTL = resourcesCacheTTL
            if cacheTTL is not None:
                logger.warn("The `resourcesCacheTTL` keyword is deprecated and will be removed in StreamFlow 0.3.0. "
                            "Use `locationsCacheTTL` instead.")
            else:
                cacheTTL = 10
        self.locationsCache: Cache = TTLCache(maxsize=cacheSize, ttl=cacheTTL)

    async def _check_effective_location(self,
                                        common_paths: MutableMapping[str, Any],
                                        effective_locations: MutableSequence[str],
                                        location: str,
                                        path: str,
                                        source_location: Optional[str] = None
                                        ) -> Tuple[MutableMapping[str, Any], MutableSequence[str]]:
        # Get all container mounts
        volumes = await self._get_volumes(location)
        for volume in volumes:
            # If path is in a persistent volume
            if path.startswith(volume['Destination']):
                if path not in common_paths:
                    common_paths[path] = []
                # Check if path is shared with another location that has been already processed
                for i, common_path in enumerate(common_paths[path]):
                    if common_path['source'] == volume['Source']:
                        # If path has already been processed, but the current location is the source, substitute it
                        if location == source_location:
                            effective_locations.remove(common_path['location'])
                            common_path['location'] = location
                            common_paths[path][i] = common_path
                            effective_locations.append(location)
                            return common_paths, effective_locations
                        # Otherwise simply skip current location
                        else:
                            return common_paths, effective_locations
                # If this is the first location encountered with the persistent path, add it to the list
                effective_locations.append(location)
                common_paths[path].append({'source': volume['Source'], 'location': location})
                return common_paths, effective_locations
        # If path is not in a persistent volume, add the current location to the list
        effective_locations.append(location)
        return common_paths, effective_locations

    async def _copy_local_to_remote(self,
                                    src: str,
                                    dst: str,
                                    locations: MutableSequence[str],
                                    read_only: bool = False) -> None:
        effective_locations = await self._get_effective_locations(locations, dst)
        copy_tasks = []
        for location in effective_locations:
            if read_only and await self._is_bind_transfer(location, src, dst):
                copy_tasks.append(asyncio.create_task(
                    self.run(
                        location=location,
                        command=["ln", "-snf", posixpath.abspath(src), dst])))
                continue
            copy_tasks.append(asyncio.create_task(
                self._copy_local_to_remote_single(
                    src=src,
                    dst=dst,
                    location=location,
                    read_only=read_only)))
        await asyncio.gather(*copy_tasks)

    async def _copy_remote_to_remote(self,
                                     src: str,
                                     dst: str,
                                     locations: MutableSequence[str],
                                     source_location: str,
                                     source_connector: Optional[Connector] = None,
                                     read_only: bool = False) -> None:
        source_connector = source_connector or self
        effective_locations = await self._get_effective_locations(locations, dst, source_location)
        non_bind_locations = []
        if read_only and source_connector == self and await self._is_bind_transfer(source_location, src, src):
            for location in effective_locations:
                if await self._is_bind_transfer(location, dst, dst):
                    await self.run(
                        location=location,
                        command=["ln", "-snf", posixpath.abspath(src), dst])
                    continue
                non_bind_locations.append(location)
            await super()._copy_remote_to_remote(
                src=src,
                dst=dst,
                locations=non_bind_locations,
                source_connector=source_connector,
                source_location=source_location,
                read_only=read_only)
        else:
            await super()._copy_remote_to_remote(
                src=src,
                dst=dst,
                locations=effective_locations,
                source_connector=source_connector,
                source_location=source_location,
                read_only=read_only)

    async def _get_effective_locations(self,
                                       locations: MutableSequence[str],
                                       dest_path: str,
                                       source_location: Optional[str] = None) -> MutableSequence[str]:
        common_paths = {}
        effective_locations = []
        for location in locations:
            common_paths, effective_locations = await self._check_effective_location(
                common_paths,
                effective_locations,
                location,
                dest_path,
                source_location)
        return effective_locations

    @abstractmethod
    async def _get_bind_mounts(self,
                               location: str) -> MutableSequence[MutableMapping[str, str]]:
        ...

    @abstractmethod
    async def _get_location(self,
                            location_name: str) -> Optional[Location]:
        ...

    @abstractmethod
    async def _get_volumes(self,
                           location: str) -> MutableSequence[MutableMapping[str, str]]:
        ...

    async def _is_bind_transfer(self,
                                location: str,
                                host_path: str,
                                instance_path: str) -> bool:
        bind_mounts = await self._get_bind_mounts(location)
        host_binds = [b['Source'] for b in bind_mounts]
        instance_binds = [b['Destination'] for b in bind_mounts]
        for host_bind in host_binds:
            if host_path.startswith(host_bind):
                for instance_bind in instance_binds:
                    if instance_path.startswith(instance_bind):
                        return True
        return False


class DockerBaseConnector(ContainerConnector, ABC):

    async def _get_bind_mounts(self,
                               location: str) -> MutableSequence[MutableMapping[str, str]]:
        return [v for v in await self._get_volumes(location) if v['Type'] == 'bind']

    async def _get_location(self,
                            location_name: str) -> Location:
        inspect_command = "".join([
            "docker ",
            "inspect ",
            "--format ",
            "'{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}' ",
            location_name])
        proc = await asyncio.create_subprocess_exec(
            *shlex.split(inspect_command),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        stdout, _ = await proc.communicate()
        return Location(name=location_name, hostname=stdout.decode().strip())

    def _get_run_command(self,
                         command: str,
                         location: str,
                         interactive: bool = False):
        return "".join([
            "docker "
            "exec ",
            "-i " if interactive else "",
            "{location} ",
            "sh -c '{command}'"
        ]).format(
            location=location,
            command=command
        )

    async def _get_volumes(self,
                           location: str) -> MutableSequence[MutableMapping[str, str]]:
        inspect_command = "".join([
            "docker ",
            "inspect ",
            "--format ",
            "'{{json .Mounts }}' ",
            location])
        proc = await asyncio.create_subprocess_exec(
            *shlex.split(inspect_command),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        stdout, _ = await proc.communicate()
        return json.loads(stdout.decode().strip())


class DockerConnector(DockerBaseConnector):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 image: str,
                 addHost: Optional[MutableSequence[str]] = None,
                 blkioWeight: Optional[int] = None,
                 blkioWeightDevice: Optional[MutableSequence[int]] = None,
                 capAdd: Optional[MutableSequence[str]] = None,
                 capDrop: Optional[MutableSequence[str]] = None,
                 cgroupParent: Optional[str] = None,
                 cidfile: Optional[str] = None,
                 containerIds: Optional[MutableSequence] = None,
                 cpuPeriod: Optional[int] = None,
                 cpuQuota: Optional[int] = None,
                 cpuRTPeriod: Optional[int] = None,
                 cpuRTRuntime: Optional[int] = None,
                 cpuShares: Optional[int] = None,
                 cpus: Optional[float] = None,
                 cpusetCpus: Optional[str] = None,
                 cpusetMems: Optional[str] = None,
                 detachKeys: Optional[str] = None,
                 device: Optional[MutableSequence[str]] = None,
                 deviceCgroupRule: Optional[MutableSequence[str]] = None,
                 deviceReadBps: Optional[MutableSequence[str]] = None,
                 deviceReadIops: Optional[MutableSequence[str]] = None,
                 deviceWriteBps: Optional[MutableSequence[str]] = None,
                 deviceWriteIops: Optional[MutableSequence[str]] = None,
                 disableContentTrust: bool = True,
                 dns: Optional[MutableSequence[str]] = None,
                 dnsOptions: Optional[MutableSequence[str]] = None,
                 dnsSearch: Optional[MutableSequence[str]] = None,
                 domainname: Optional[str] = None,
                 entrypoint: Optional[str] = None,
                 env: Optional[MutableSequence[str]] = None,
                 envFile: Optional[MutableSequence[str]] = None,
                 expose: Optional[MutableSequence[str]] = None,
                 gpus: Optional[MutableSequence[str]] = None,
                 groupAdd: Optional[MutableSequence[str]] = None,
                 healthCmd: Optional[str] = None,
                 healthInterval: Optional[str] = None,
                 healthRetries: Optional[int] = None,
                 healthStartPeriod: Optional[str] = None,
                 healthTimeout: Optional[str] = None,
                 hostname: Optional[str] = None,
                 init: bool = True,
                 ip: Optional[str] = None,
                 ip6: Optional[str] = None,
                 ipc: Optional[str] = None,
                 isolation: Optional[str] = None,
                 kernelMemory: Optional[int] = None,
                 label: Optional[MutableSequence[str]] = None,
                 labelFile: Optional[MutableSequence[str]] = None,
                 link: Optional[MutableSequence[str]] = None,
                 linkLocalIP: Optional[MutableSequence[str]] = None,
                 locationsCacheSize: int = None,
                 locationsCacheTTL: int = None,
                 logDriver: Optional[str] = None,
                 logOpts: Optional[MutableSequence[str]] = None,
                 macAddress: Optional[str] = None,
                 memory: Optional[int] = None,
                 memoryReservation: Optional[int] = None,
                 memorySwap: Optional[int] = None,
                 memorySwappiness: Optional[int] = None,
                 mount: Optional[MutableSequence[str]] = None,
                 network: Optional[MutableSequence[str]] = None,
                 networkAlias: Optional[MutableSequence[str]] = None,
                 noHealthcheck: bool = False,
                 oomKillDisable: bool = False,
                 oomScoreAdj: Optional[int] = None,
                 pid: Optional[str] = None,
                 pidsLimit: Optional[int] = None,
                 privileged: bool = False,
                 publish: Optional[MutableSequence[str]] = None,
                 publishAll: bool = False,
                 readOnly: bool = False,
                 replicas: int = 1,
                 resourcesCacheSize: int = None,
                 resourcesCacheTTL: int = None,
                 restart: Optional[str] = None,
                 rm: bool = True,
                 runtime: Optional[str] = None,
                 securityOpts: Optional[MutableSequence[str]] = None,
                 shmSize: Optional[int] = None,
                 sigProxy: bool = True,
                 stopSignal: Optional[str] = None,
                 stopTimeout: Optional[int] = None,
                 storageOpts: Optional[MutableSequence[str]] = None,
                 sysctl: Optional[MutableSequence[str]] = None,
                 tmpfs: Optional[MutableSequence[str]] = None,
                 transferBufferSize: int = 2 ** 16,
                 ulimit: Optional[MutableSequence[str]] = None,
                 user: Optional[str] = None,
                 userns: Optional[str] = None,
                 uts: Optional[str] = None,
                 volume: Optional[MutableSequence[str]] = None,
                 volumeDriver: Optional[str] = None,
                 volumesFrom: Optional[MutableSequence[str]] = None,
                 workdir: Optional[str] = None):
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            transferBufferSize=transferBufferSize,
            locationsCacheSize=locationsCacheSize,
            locationsCacheTTL=locationsCacheTTL,
            resourcesCacheSize=resourcesCacheSize,
            resourcesCacheTTL=resourcesCacheTTL)
        self.image: str = image
        self.addHost: Optional[MutableSequence[str]] = addHost
        self.blkioWeight: Optional[int] = blkioWeight
        self.blkioWeightDevice: Optional[MutableSequence[int]] = blkioWeightDevice
        self.capAdd: Optional[MutableSequence[str]] = capAdd
        self.capDrop: Optional[MutableSequence[str]] = capDrop
        self.cgroupParent: Optional[str] = cgroupParent
        self.cidfile: Optional[str] = cidfile
        self.containerIds: MutableSequence[str] = containerIds or []
        self.cpuPeriod: Optional[int] = cpuPeriod
        self.cpuQuota: Optional[int] = cpuQuota
        self.cpuRTPeriod: Optional[int] = cpuRTPeriod
        self.cpuRTRuntime: Optional[int] = cpuRTRuntime
        self.cpuShares: Optional[int] = cpuShares
        self.cpus: Optional[float] = cpus
        self.cpusetCpus: Optional[str] = cpusetCpus
        self.cpusetMems: Optional[str] = cpusetMems
        self.detachKeys: Optional[str] = detachKeys
        self.device: Optional[MutableSequence[str]] = device
        self.deviceCgroupRule: Optional[MutableSequence[str]] = deviceCgroupRule
        self.deviceReadBps: Optional[MutableSequence[str]] = deviceReadBps
        self.deviceReadIops: Optional[MutableSequence[str]] = deviceReadIops
        self.deviceWriteBps: Optional[MutableSequence[str]] = deviceWriteBps
        self.deviceWriteIops: Optional[MutableSequence[str]] = deviceWriteIops
        self.disableContentTrust: bool = disableContentTrust
        self.dns: Optional[MutableSequence[str]] = dns
        self.dnsOptions: Optional[MutableSequence[str]] = dnsOptions
        self.dnsSearch: Optional[MutableSequence[str]] = dnsSearch
        self.domainname: Optional[str] = domainname
        self.entrypoint: Optional[str] = entrypoint
        self.env: Optional[MutableSequence[str]] = env
        self.envFile: Optional[MutableSequence[str]] = envFile
        self.expose: Optional[MutableSequence[str]] = expose
        self.gpus: Optional[MutableSequence[str]] = gpus
        self.groupAdd: Optional[MutableSequence[str]] = groupAdd
        self.healthCmd: Optional[str] = healthCmd
        self.healthInterval: Optional[str] = healthInterval
        self.healthRetries: Optional[int] = healthRetries
        self.healthStartPeriod: Optional[str] = healthStartPeriod
        self.healthTimeout: Optional[str] = healthTimeout
        self.hostname: Optional[str] = hostname
        self.init: bool = init
        self.ip: Optional[str] = ip
        self.ip6: Optional[str] = ip6
        self.ipc: Optional[str] = ipc
        self.isolation: Optional[str] = isolation
        self.kernelMemory: Optional[int] = kernelMemory
        self.label: Optional[MutableSequence[str]] = label
        self.labelFile: Optional[MutableSequence[str]] = labelFile
        self.link: Optional[MutableSequence[str]] = link
        self.linkLocalIP: Optional[MutableSequence[str]] = linkLocalIP
        self.logDriver: Optional[str] = logDriver
        self.logOpts: Optional[MutableSequence[str]] = logOpts
        self.macAddress: Optional[str] = macAddress
        self.memory: Optional[int] = memory
        self.memoryReservation: Optional[int] = memoryReservation
        self.memorySwap: Optional[int] = memorySwap
        self.memorySwappiness: Optional[int] = memorySwappiness
        self.mount: Optional[MutableSequence[str]] = mount
        self.network: Optional[MutableSequence[str]] = network
        self.networkAlias: Optional[MutableSequence[str]] = networkAlias
        self.noHealthcheck: bool = noHealthcheck
        self.oomKillDisable: bool = oomKillDisable
        self.oomScoreAdj: Optional[int] = oomScoreAdj
        self.pid: Optional[str] = pid
        self.pidsLimit: Optional[int] = pidsLimit
        self.privileged: bool = privileged
        self.publish: Optional[MutableSequence[str]] = publish
        self.publishAll: bool = publishAll
        self.readOnly: bool = readOnly
        self.replicas: int = replicas
        self.restart: Optional[str] = restart
        self.rm: bool = rm
        self.runtime: Optional[str] = runtime
        self.securityOpts: Optional[MutableSequence[str]] = securityOpts
        self.shmSize: Optional[int] = shmSize
        self.sigProxy: bool = sigProxy
        self.stopSignal: Optional[str] = stopSignal
        self.stopTimeout: Optional[int] = stopTimeout
        self.storageOpts: Optional[MutableSequence[str]] = storageOpts
        self.sysctl: Optional[MutableSequence[str]] = sysctl
        self.tmpfs: Optional[MutableSequence[str]] = tmpfs
        self.ulimit: Optional[MutableSequence[str]] = ulimit
        self.user: Optional[str] = user
        self.userns: Optional[str] = userns
        self.uts: Optional[str] = uts
        self.volume: Optional[MutableSequence[str]] = volume
        self.volumeDriver: Optional[str] = volumeDriver
        self.volumesFrom: Optional[MutableSequence[str]] = volumesFrom
        self.workdir: Optional[str] = workdir

    async def deploy(self, external: bool) -> None:
        if not external:
            # Pull image if it doesn't exist
            if not await _exists_docker_image(self.image):
                await _pull_docker_image(self.image)
            # Deploy the Docker container
            for _ in range(0, self.replicas):
                deploy_command = "".join([
                    "docker ",
                    "run "
                    "--detach "
                    "--interactive "
                    "{addHost}"
                    "{blkioWeight}"
                    "{blkioWeightDevice}"
                    "{capAdd}"
                    "{capDrop}"
                    "{cgroupParent}"
                    "{cidfile}"
                    "{cpuPeriod}"
                    "{cpuQuota}"
                    "{cpuRTPeriod}"
                    "{cpuRTRuntime}"
                    "{cpuShares}"
                    "{cpus}"
                    "{cpusetCpus}"
                    "{cpusetMems}"
                    "{detachKeys}"
                    "{device}"
                    "{deviceCgroupRule}"
                    "{deviceReadBps}"
                    "{deviceReadIops}"
                    "{deviceWriteBps}"
                    "{deviceWriteIops}"
                    "{disableContentTrust}"
                    "{dns}"
                    "{dnsOptions}"
                    "{dnsSearch}"
                    "{domainname}"
                    "{entrypoint}"
                    "{env}"
                    "{envFile}"
                    "{expose}"
                    "{gpus}"
                    "{groupAdd}"
                    "{healthCmd}"
                    "{healthInterval}"
                    "{healthRetries}"
                    "{healthStartPeriod}"
                    "{healthTimeout}"
                    "{hostname}"
                    "{init}"
                    "{ip}"
                    "{ip6}"
                    "{ipc}"
                    "{isolation}"
                    "{kernelMemory}"
                    "{label}"
                    "{labelFile}"
                    "{link}"
                    "{linkLocalIP}"
                    "{logDriver}"
                    "{logOpts}"
                    "{macAddress}"
                    "{memory}"
                    "{memoryReservation}"
                    "{memorySwap}"
                    "{memorySwappiness}"
                    "{mount}"
                    "{network}"
                    "{networkAlias}"
                    "{noHealthcheck}"
                    "{oomKillDisable}"
                    "{oomScoreAdj}"
                    "{pid}"
                    "{pidsLimit}"
                    "{privileged}"
                    "{publish}"
                    "{publishAll}"
                    "{readOnly}"
                    "{restart}"
                    "{rm}"
                    "{runtime}"
                    "{securityOpts}"
                    "{shmSize}"
                    "{sigProxy}"
                    "{stopSignal}"
                    "{stopTimeout}"
                    "{storageOpts}"
                    "{sysctl}"
                    "{tmpfs}"
                    "{ulimit}"
                    "{user}"
                    "{userns}"
                    "{uts}"
                    "{volume}"
                    "{volumeDriver}"
                    "{volumesFrom}"
                    "{workdir}"
                    "{image}"
                ]).format(
                    addHost=self.get_option("add-host", self.addHost),
                    blkioWeight=self.get_option("blkio-weight", self.addHost),
                    blkioWeightDevice=self.get_option("blkio-weight-device", self.blkioWeightDevice),
                    capAdd=self.get_option("cap-add", self.capAdd),
                    capDrop=self.get_option("cap-drop", self.capDrop),
                    cgroupParent=self.get_option("cgroup-parent", self.cgroupParent),
                    cidfile=self.get_option("cidfile", self.cidfile),
                    cpuPeriod=self.get_option("cpu-period", self.cpuPeriod),
                    cpuQuota=self.get_option("cpu-quota", self.cpuQuota),
                    cpuRTPeriod=self.get_option("cpu-rt-period", self.cpuRTPeriod),
                    cpuRTRuntime=self.get_option("cpu-rt-runtime", self.cpuRTRuntime),
                    cpuShares=self.get_option("cpu-shares", self.cpuShares),
                    cpus=self.get_option("cpus", self.cpus),
                    cpusetCpus=self.get_option("cpuset-cpus", self.cpusetCpus),
                    cpusetMems=self.get_option("cpuset-mems", self.cpusetMems),
                    detachKeys=self.get_option("detach-keys", self.detachKeys),
                    device=self.get_option("device", self.device),
                    deviceCgroupRule=self.get_option("device-cgroup-rule", self.deviceCgroupRule),
                    deviceReadBps=self.get_option("device-read-bps", self.deviceReadBps),
                    deviceReadIops=self.get_option("device-read-iops", self.deviceReadIops),
                    deviceWriteBps=self.get_option("device-write-bps", self.deviceWriteBps),
                    deviceWriteIops=self.get_option("device-write-iops", self.deviceWriteIops),
                    disableContentTrust="--disable-content-trust={disableContentTrust} ".format(
                        disableContentTrust="true" if self.disableContentTrust else "false"),
                    dns=self.get_option("dns", self.dns),
                    dnsOptions=self.get_option("dns-option", self.dnsOptions),
                    dnsSearch=self.get_option("dns-search", self.dnsSearch),
                    domainname=self.get_option("domainname", self.domainname),
                    entrypoint=self.get_option("entrypoint", self.entrypoint),
                    env=self.get_option("env", self.env),
                    envFile=self.get_option("env-file", self.envFile),
                    expose=self.get_option("expose", self.expose),
                    gpus=self.get_option("gpus", self.gpus),
                    groupAdd=self.get_option("group-add", self.groupAdd),
                    healthCmd=self.get_option("health-cmd", self.healthCmd),
                    healthInterval=self.get_option("health-interval", self.healthInterval),
                    healthRetries=self.get_option("health-retries", self.healthRetries),
                    healthStartPeriod=self.get_option("health-start-period", self.healthStartPeriod),
                    healthTimeout=self.get_option("health-timeout", self.healthTimeout),
                    hostname=self.get_option("hostname", self.hostname),
                    init=self.get_option("init", self.init),
                    ip=self.get_option("ip", self.ip),
                    ip6=self.get_option("ip6", self.ip6),
                    ipc=self.get_option("ipc", self.ipc),
                    isolation=self.get_option("isolation", self.isolation),
                    kernelMemory=self.get_option("kernel-memory", self.kernelMemory),
                    label=self.get_option("label", self.label),
                    labelFile=self.get_option("label-file", self.labelFile),
                    link=self.get_option("link", self.link),
                    linkLocalIP=self.get_option("link-local-ip", self.linkLocalIP),
                    logDriver=self.get_option("log-driver", self.logDriver),
                    logOpts=self.get_option("log-opt", self.logOpts),
                    macAddress=self.get_option("mac-address", self.macAddress),
                    memory=self.get_option("memory", self.memory),
                    memoryReservation=self.get_option("memory-reservation", self.memoryReservation),
                    memorySwap=self.get_option("memory-swap", self.memorySwap),
                    memorySwappiness=self.get_option("memory-swappiness", self.memorySwappiness),
                    mount=self.get_option("mount", self.mount),
                    network=self.get_option("network", self.network),
                    networkAlias=self.get_option("network-alias", self.networkAlias),
                    noHealthcheck=self.get_option("no-healthcheck", self.noHealthcheck),
                    oomKillDisable=self.get_option("oom-kill-disable", self.oomKillDisable),
                    oomScoreAdj=self.get_option("oom-score-adj", self.oomScoreAdj),
                    pid=self.get_option("pid", self.pid),
                    pidsLimit=self.get_option("pids-limit", self.pidsLimit),
                    privileged=self.get_option("privileged", self.privileged),
                    publish=self.get_option("publish", self.publish),
                    publishAll=self.get_option("publish-all", self.publishAll),
                    readOnly=self.get_option("read-only", self.readOnly),
                    restart=self.get_option("restart", self.restart),
                    rm=self.get_option("rm", self.rm),
                    runtime=self.get_option("runtime", self.runtime),
                    securityOpts=self.get_option("security-opt", self.securityOpts),
                    shmSize=self.get_option("shm-size", self.shmSize),
                    sigProxy="--sig-proxy={sigProxy} ".format(sigProxy="true" if self.sigProxy else "false"),
                    stopSignal=self.get_option("stop-signal", self.stopSignal),
                    stopTimeout=self.get_option("stop-timeout", self.stopTimeout),
                    storageOpts=self.get_option("storage-opt", self.storageOpts),
                    sysctl=self.get_option("sysctl", self.sysctl),
                    tmpfs=self.get_option("tmpfs", self.tmpfs),
                    ulimit=self.get_option("ulimit", self.ulimit),
                    user=self.get_option("user", self.user),
                    userns=self.get_option("userns", self.userns),
                    uts=self.get_option("uts", self.uts),
                    volume=self.get_option("volume", self.volume),
                    volumeDriver=self.get_option("volume-driver", self.volumeDriver),
                    volumesFrom=self.get_option("volumes-from", self.volumesFrom),
                    workdir=self.get_option("workdir", self.workdir),
                    image=self.image
                )
                logger.debug("Executing command {command}".format(command=deploy_command))
                proc = await asyncio.create_subprocess_exec(
                    *shlex.split(deploy_command),
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE)
                stdout, stderr = await proc.communicate()
                if proc.returncode == 0:
                    self.containerIds.append(stdout.decode().strip())
                else:
                    raise WorkflowExecutionException(stderr.decode().strip())

    @cachedmethod(lambda self: self.locationsCache)
    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        return {container_id: await self._get_location(container_id) for container_id in self.containerIds}

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'docker.json'))

    async def undeploy(self, external: bool) -> None:
        if not external and self.containerIds:
            for container_id in self.containerIds:
                undeploy_command = "".join([
                    "docker ",
                    "stop "
                    "{containerId}"
                ]).format(
                    containerId=container_id
                )
                logger.debug("Executing command {command}".format(command=undeploy_command))
                proc = await asyncio.create_subprocess_exec(
                    *shlex.split(undeploy_command),
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE)
                await proc.wait()
            self.containerIds = []


class DockerComposeConnector(DockerBaseConnector):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 files: MutableSequence[str],
                 projectName: Optional[str] = None,
                 verbose: Optional[bool] = False,
                 logLevel: Optional[str] = None,
                 noAnsi: Optional[bool] = False,
                 host: Optional[str] = None,
                 tls: Optional[bool] = False,
                 tlscacert: Optional[str] = None,
                 tlscert: Optional[str] = None,
                 tlskey: Optional[str] = None,
                 tlsverify: Optional[bool] = False,
                 skipHostnameCheck: Optional[bool] = False,
                 projectDirectory: Optional[str] = None,
                 compatibility: Optional[bool] = False,
                 noDeps: Optional[bool] = False,
                 forceRecreate: Optional[bool] = False,
                 alwaysRecreateDeps: Optional[bool] = False,
                 noRecreate: Optional[bool] = False,
                 noBuild: Optional[bool] = False,
                 noStart: Optional[bool] = False,
                 build: Optional[bool] = False,
                 timeout: Optional[int] = None,
                 transferBufferSize: int = 2 ** 16,
                 renewAnonVolumes: Optional[bool] = False,
                 removeOrphans: Optional[bool] = False,
                 removeVolumes: Optional[bool] = False,
                 locationsCacheSize: int = None,
                 locationsCacheTTL: int = None,
                 resourcesCacheSize: int = None,
                 resourcesCacheTTL: int = None) -> None:
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            transferBufferSize=transferBufferSize,
            locationsCacheSize=locationsCacheSize,
            locationsCacheTTL=locationsCacheTTL,
            resourcesCacheSize=resourcesCacheSize,
            resourcesCacheTTL=resourcesCacheTTL)
        self.files = [os.path.join(context.config_dir, file) for file in files]
        self.projectName = projectName
        self.verbose = verbose
        self.logLevel = logLevel
        self.noAnsi = noAnsi
        self.host = host
        self.noDeps = noDeps
        self.forceRecreate = forceRecreate
        self.alwaysRecreateDeps = alwaysRecreateDeps
        self.noRecreate = noRecreate
        self.noBuild = noBuild
        self.noStart = noStart
        self.build = build
        self.renewAnonVolumes = renewAnonVolumes
        self.removeOrphans = removeOrphans
        self.removeVolumes = removeVolumes
        self.skipHostnameCheck = skipHostnameCheck
        self.projectDirectory = projectDirectory
        self.compatibility = compatibility
        self.timeout = timeout
        self.tls = tls
        self.tlscacert = tlscacert
        self.tlscert = tlscert
        self.tlskey = tlskey
        self.tlsverify = tlsverify

    def base_command(self) -> str:
        return "".join([
            "docker-compose ",
            "{files}",
            "{projectName}",
            "{verbose}",
            "{logLevel}",
            "{noAnsi}",
            "{host}",
            "{tls}",
            "{tlscacert}",
            "{tlscert}",
            "{tlskey}",
            "{tlsverify}",
            "{skipHostnameCheck}",
            "{projectDirectory}",
            "{compatibility} "
        ]).format(
            files=self.get_option("file", self.files),
            projectName=self.get_option("project-name", self.projectName),
            verbose=self.get_option("verbose", self.verbose),
            logLevel=self.get_option("log-level", self.logLevel),
            noAnsi=self.get_option("no-ansi", self.noAnsi),
            host=self.get_option("host", self.host),
            tls=self.get_option("tls", self.tls),
            tlscacert=self.get_option("tlscacert", self.tlscacert),
            tlscert=self.get_option("tlscert", self.tlscert),
            tlskey=self.get_option("tlskey", self.tlskey),
            tlsverify=self.get_option("tlsverify", self.tlsverify),
            skipHostnameCheck=self.get_option("skip-hostname-check", self.skipHostnameCheck),
            projectDirectory=self.get_option("project-directory", self.projectDirectory),
            compatibility=self.get_option("compatibility", self.compatibility)
        )

    async def deploy(self, external: bool) -> None:
        if not external:
            deploy_command = self.base_command() + "".join([
                "up ",
                "--detach ",
                "{noDeps}",
                "{forceRecreate}",
                "{alwaysRecreateDeps}",
                "{noRecreate}",
                "{noBuild}",
                "{noStart}"
            ]).format(
                noDeps=self.get_option("no-deps ", self.noDeps),
                forceRecreate=self.get_option("force-recreate", self.forceRecreate),
                alwaysRecreateDeps=self.get_option("always-recreate-deps", self.alwaysRecreateDeps),
                noRecreate=self.get_option("no-recreate", self.noRecreate),
                noBuild=self.get_option("no-build", self.noBuild),
                noStart=self.get_option("no-start", self.noStart)
            )
            logger.debug("Executing command {command}".format(command=deploy_command))
            proc = await asyncio.create_subprocess_exec(*shlex.split(deploy_command))
            await proc.wait()

    @cachedmethod(lambda self: self.locationsCache)
    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        ps_command = self.base_command() + "".join([
            "ps ",
            service or ""])
        logger.debug("Executing command {command}".format(command=ps_command))
        proc = await asyncio.create_subprocess_exec(
            *shlex.split(ps_command),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        stdout, _ = await proc.communicate()
        lines = (line for line in stdout.decode().strip().split('\n'))
        locations = {}
        for line in lines:
            if line.startswith("---------"):
                break
        for line in lines:
            location_name = line.split()[0].strip()
            locations[location_name] = await self._get_location(location_name)
        return locations

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'docker-compose.json'))

    async def undeploy(self, external: bool) -> None:
        if not external:
            undeploy_command = self.base_command() + "".join([
                "down ",
                "{removeVolumes}"
            ]).format(
                removeVolumes=self.get_option("volumes", self.removeVolumes)
            )
            logger.debug("Executing command {command}".format(command=undeploy_command))
            proc = await asyncio.create_subprocess_exec(*shlex.split(undeploy_command))
            await proc.wait()


class SingularityBaseConnector(ContainerConnector, ABC):

    async def _get_bind_mounts(self, location: str) -> MutableSequence[MutableMapping[str, str]]:
        return await self._get_volumes(location)

    async def _get_location(self, location_name: str) -> Optional[Location]:
        inspect_command = "singularity instance list --json"
        proc = await asyncio.create_subprocess_exec(
            *shlex.split(inspect_command),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        stdout, _ = await proc.communicate()
        if stdout:
            json_out = json.loads(stdout)
            for instance in json_out["instances"]:
                if instance["instance"] == location_name:
                    return Location(name=location_name, hostname=instance["ip"])
        else:
            return None

    def _get_run_command(self,
                         command: str,
                         location: str,
                         interactive: bool = False):
        return "".join([
            "singularity "
            "exec ",
            "instance://{location} "
            "sh -c '{command}'"
        ]).format(
            location=location,
            command=command
        )

    async def _get_volumes(self,
                           location: str) -> MutableSequence[MutableMapping[str, str]]:
        bind_mounts, _ = await self.run(
            location=location,
            command=["cat", "/proc/1/mountinfo", "|", "awk", "'{print $9,$4,$5}'"],
            capture_output=True)
        # Exclude `overlay` and `tmpfs` mounts
        return [{'Source': line.split()[1], 'Destination': line.split()[2]} for line in bind_mounts.splitlines()
                if line.split()[0] not in ['tmpfs', 'overlay']]


class SingularityConnector(SingularityBaseConnector):

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 image: str,
                 transferBufferSize: int = 2 ** 16,
                 addCaps: Optional[str] = None,
                 allowSetuid: bool = False,
                 applyCgroups: Optional[str] = None,
                 bind: Optional[MutableSequence[str]] = None,
                 boot: bool = False,
                 cleanenv: bool = False,
                 contain: bool = False,
                 containall: bool = False,
                 disableCache: bool = False,
                 dns: Optional[str] = None,
                 dropCaps: Optional[str] = None,
                 env: Optional[MutableSequence[str]] = None,
                 envFile: Optional[str] = None,
                 fakeroot: bool = False,
                 fusemount: Optional[MutableSequence[str]] = None,
                 home: Optional[str] = None,
                 hostname: Optional[str] = None,
                 instanceNames: Optional[MutableSequence[str]] = None,
                 keepPrivs: bool = False,
                 locationsCacheSize: int = None,
                 locationsCacheTTL: int = None,
                 net: bool = False,
                 network: Optional[str] = None,
                 networkArgs: Optional[MutableSequence[str]] = None,
                 noHome: bool = False,
                 noInit: bool = False,
                 noMount: Optional[MutableSequence[str]] = None,
                 noPrivs: bool = False,
                 noUmask: bool = False,
                 nohttps: bool = False,
                 nv: bool = False,
                 overlay: Optional[MutableSequence[str]] = None,
                 pemPath: Optional[str] = None,
                 pidFile: Optional[str] = None,
                 replicas: int = 1,
                 resourcesCacheSize: int = None,
                 resourcesCacheTTL: int = None,
                 rocm: bool = False,
                 scratch: Optional[MutableSequence[str]] = None,
                 security: Optional[MutableSequence[str]] = None,
                 userns: bool = False,
                 uts: bool = False,
                 workdir: Optional[str] = None,
                 writable: bool = False,
                 writableTmpfs: bool = False):
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            transferBufferSize=transferBufferSize,
            locationsCacheSize=locationsCacheSize,
            locationsCacheTTL=locationsCacheTTL,
            resourcesCacheSize=resourcesCacheSize,
            resourcesCacheTTL=resourcesCacheTTL)
        self.image: str = image
        self.addCaps: Optional[str] = addCaps
        self.allowSetuid: bool = allowSetuid
        self.applyCgroups: Optional[str] = applyCgroups
        self.bind: Optional[MutableSequence[str]] = bind
        self.boot: bool = boot
        self.cleanenv: bool = cleanenv
        self.contain: bool = contain
        self.containall: bool = containall
        self.disableCache: bool = disableCache
        self.dns: Optional[str] = dns
        self.dropCaps: Optional[str] = dropCaps
        self.env: Optional[MutableSequence[str]] = env
        self.envFile: Optional[str] = envFile
        self.fakeroot: bool = fakeroot
        self.fusemount: Optional[MutableSequence[str]] = fusemount
        self.home: Optional[str] = home
        self.hostname: Optional[str] = hostname
        self.instanceNames: MutableSequence[str] = instanceNames or []
        self.keepPrivs: bool = keepPrivs
        self.net: bool = net
        self.network: Optional[str] = network
        self.networkArgs: Optional[MutableSequence[str]] = networkArgs
        self.noHome: bool = noHome
        self.noInit: bool = noInit
        self.noMount: Optional[MutableSequence[str]] = noMount or []
        self.noPrivs: bool = noPrivs
        self.noUmask: bool = noUmask
        self.nohttps: bool = nohttps
        self.nv: bool = nv
        self.overlay: Optional[MutableSequence[str]] = overlay
        self.pemPath: Optional[str] = pemPath
        self.pidFile: Optional[str] = pidFile
        self.replicas: int = replicas
        self.rocm: bool = rocm
        self.scratch: Optional[MutableSequence[str]] = scratch
        self.security: Optional[MutableSequence[str]] = security
        self.userns: bool = userns
        self.uts: bool = uts
        self.workdir: Optional[str] = workdir
        self.writable: bool = writable
        self.writableTmpfs: bool = writableTmpfs

    async def deploy(self, external: bool) -> None:
        if not external:
            for _ in range(0, self.replicas):
                instance_name = utils.random_name()
                deploy_command = "".join([
                    "singularity ",
                    "instance "
                    "start "
                    "{addCaps}"
                    "{allowSetuid}"
                    "{applyCgroups}"
                    "{bind}"
                    "{boot}"
                    "{cleanenv}"
                    "{contain}"
                    "{containall}"
                    "{disableCache}"
                    "{dns}"
                    "{dropCaps}"
                    "{env}"
                    "{envFile}"
                    "{fakeroot}"
                    "{fusemount}"
                    "{home}"
                    "{hostname}"
                    "{keepPrivs}"
                    "{net}"
                    "{network}"
                    "{networkArgs}"
                    "{noHome}"
                    "{noInit}"
                    "{noMount}"
                    "{noPrivs}"
                    "{noUmask}"
                    "{nohttps}"
                    "{nv}"
                    "{overlay}"
                    "{pemPath}"
                    "{pidFile}"
                    "{rocm}"
                    "{scratch}"
                    "{security}"
                    "{userns}"
                    "{uts}"
                    "{workdir}"
                    "{writable}"
                    "{writableTmpfs}"
                    "{image} "
                    "{name}"
                ]).format(
                    addCaps=self.get_option("add-caps", self.addCaps),
                    allowSetuid=self.get_option("allow-setuid", self.allowSetuid),
                    applyCgroups=self.get_option("apply-cgroups", self.applyCgroups),
                    bind=self.get_option("bind", self.bind),
                    boot=self.get_option("boot", self.boot),
                    cleanenv=self.get_option("cleanenv", self.cleanenv),
                    contain=self.get_option("contain", self.contain),
                    containall=self.get_option("containall", self.containall),
                    disableCache=self.get_option("disable-cache", self.disableCache),
                    dns=self.get_option("dns", self.dns),
                    dropCaps=self.get_option("drop-caps", self.dropCaps),
                    env=self.get_option("env", self.env),
                    envFile=self.get_option("env-file", self.envFile),
                    fakeroot=self.get_option("fakeroot", self.fakeroot),
                    fusemount=self.get_option("fusemount", self.fusemount),
                    home=self.get_option("home", self.home),
                    hostname=self.get_option("hostname", self.hostname),
                    keepPrivs=self.get_option("keep-privs", self.keepPrivs),
                    net=self.get_option("net", self.net),
                    network=self.get_option("network", self.network),
                    networkArgs=self.get_option("network-args", self.networkArgs),
                    noHome=self.get_option("no-home", self.noHome),
                    noInit=self.get_option("no-init", self.noInit),
                    noMount=self.get_option("no-mount", self.noMount),
                    noPrivs=self.get_option("no-privs", self.noPrivs),
                    noUmask=self.get_option("no-umask", self.noUmask),
                    nohttps=self.get_option("nohttps", self.nohttps),
                    nv=self.get_option("nv", self.nv),
                    overlay=self.get_option("overlay", self.overlay),
                    pemPath=self.get_option("pem-path", self.pemPath),
                    pidFile=self.get_option("pid-file", self.pidFile),
                    rocm=self.get_option("rocm", self.rocm),
                    scratch=self.get_option("scratch", self.scratch),
                    security=self.get_option("security", self.security),
                    userns=self.get_option("userns", self.userns),
                    uts=self.get_option("uts", self.uts),
                    workdir=self.get_option("workdir", self.workdir),
                    writable=self.get_option("writable", self.writable),
                    writableTmpfs=self.get_option("writable-tmpfs", self.writableTmpfs),
                    image=self.image,
                    name=instance_name
                )
                logger.debug("Executing command {command}".format(command=deploy_command))
                proc = await asyncio.create_subprocess_exec(
                    *shlex.split(deploy_command),
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE)
                stdout, stderr = await proc.communicate()
                if proc.returncode == 0:
                    self.instanceNames.append(instance_name)
                else:
                    raise WorkflowExecutionException(stderr.decode().strip())

    @cachedmethod(lambda self: self.locationsCache)
    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        location_tasks = {}
        for instance_name in self.instanceNames:
            location_tasks[instance_name] = asyncio.create_task(self._get_location(instance_name))
        return {k: v for k, v in zip(location_tasks.keys(), await asyncio.gather(*location_tasks.values()))
                if v is not None}

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'singularity.json'))

    async def undeploy(self, external: bool) -> None:
        if not external and self.instanceNames:
            for instance_name in self.instanceNames:
                undeploy_command = "".join([
                    "singularity ",
                    "instance "
                    "stop "
                    "{containerId}"
                ]).format(
                    containerId=instance_name
                )
                logger.debug("Executing command {command}".format(command=undeploy_command))
                proc = await asyncio.create_subprocess_exec(
                    *shlex.split(undeploy_command),
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE)
                await proc.wait()
            self.instanceNames = []



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING, Type, cast

from streamflow.core import utils
from streamflow.core.config import Config, SchemaEntity
from streamflow.core.context import StreamFlowContext
from streamflow.core.deployment import Connector, Target
from streamflow.core.persistence import DatabaseLoadingContext, PersistableEntity

if TYPE_CHECKING:
    from streamflow.core.workflow import Job, Status, Token
    from typing import MutableSequence, MutableMapping, Optional


class Hardware(object):
    def __init__(self,
                 cores: float = 0.0,
                 memory: float = 0.0,
                 input_directory: float = 0.0,
                 output_directory: float = 0.0,
                 tmp_directory: float = 0.0):
        self.cores: float = cores
        self.memory: float = memory
        self.input_directory: float = input_directory
        self.output_directory: float = output_directory
        self.tmp_directory: float = tmp_directory

    def __add__(self, other):
        if not isinstance(other, Hardware):
            return NotImplemented
        return self.__class__(**{k: vars(self).get(k, 0.0) + vars(other).get(k, 0.0)
                                 for k in vars(self).keys()})

    def __sub__(self, other):
        if not isinstance(other, Hardware):
            return NotImplemented
        return self.__class__(**{k: vars(self).get(k, 0.0) - vars(other).get(k, 0.0)
                                 for k in vars(self).keys()})

    def __ge__(self, other):
        if not isinstance(other, Hardware):
            return NotImplemented
        return all((vars(self).get(k, 0.0) >= vars(other).get(k, 0.0)
                    for k in set().union(vars(self).keys(), vars(other).keys())))

    def __le__(self, other):
        if not isinstance(other, Hardware):
            return NotImplemented
        return all((vars(self).get(k, 0.0) <= vars(other).get(k, 0.0)
                    for k in set().union(vars(self).keys(), vars(other).keys())))


class HardwareRequirement(ABC):

    @classmethod
    @abstractmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext):
        ...

    @abstractmethod
    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    def eval(self, inputs: MutableMapping[str, Token]) -> Hardware:
        ...

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   row: MutableMapping[str, Any],
                   loading_context: DatabaseLoadingContext):
        type = utils.get_class_from_name(row['type'])
        return await cast(Type[HardwareRequirement], type)._load(context, row['params'], loading_context)

    async def save(self, context: StreamFlowContext):
        return {'type': utils.get_class_fullname(type(self)),
                'params': await self._save_additional_params(context)}


class JobAllocation(object):
    __slots__ = ('job', 'target', 'locations', 'status', 'hardware')

    def __init__(self,
                 job: str,
                 target: Target,
                 locations: MutableSequence[str],
                 status: Status,
                 hardware: Hardware):
        self.job: str = job
        self.target: Target = target
        self.locations: MutableSequence[str] = locations
        self.status: Status = status
        self.hardware: Hardware = hardware


class Location(object):
    __slots__ = ('name', 'hostname', 'hardware', 'slots')

    def __init__(self,
                 name: str,
                 hostname: str,
                 slots: int = 1,
                 hardware: Optional[Hardware] = None):
        self.name: str = name
        self.hostname: str = hostname
        self.slots: int = slots
        self.hardware: Optional[Hardware] = hardware


class LocationAllocation(object):
    __slots__ = ('name', 'deployment', 'jobs')

    def __init__(self,
                 name: str,
                 deployment: str):
        self.name: str = name
        self.deployment: str = deployment
        self.jobs: MutableSequence[str] = []


class PolicyConfig(Config, PersistableEntity):

    async def save(self, context: StreamFlowContext):
        # TODO
        return self.config


class Policy(SchemaEntity):

    @abstractmethod
    async def get_location(self,
                           context: StreamFlowContext,
                           job: Job,
                           deployment: str,
                           hardware_requirement: Hardware,
                           available_locations: MutableMapping[str, Location],
                           jobs: MutableMapping[str, JobAllocation],
                           locations: MutableMapping[str, LocationAllocation]) -> Optional[str]: ...


class Scheduler(SchemaEntity):

    def __init__(self, context: StreamFlowContext):
        self.context: StreamFlowContext = context
        self.job_allocations: MutableMapping[str, JobAllocation] = {}
        self.location_allocations: MutableMapping[str, LocationAllocation] = {}

    @abstractmethod
    async def close(self):
        ...

    def get_allocation(self, job_name: str) -> Optional[JobAllocation]:
        return self.job_allocations.get(job_name)

    def get_hardware(self, job_name: str) -> Optional[Hardware]:
        allocation = self.get_allocation(job_name)
        return allocation.hardware if allocation else None

    def get_connector(self, job_name: str) -> Optional[Connector]:
        allocation = self.get_allocation(job_name)
        return self.context.deployment_manager.get_connector(allocation.target.deployment.name) if allocation else None

    def get_locations(self,
                      job_name: str,
                      statuses: Optional[MutableSequence[Status]] = None) -> MutableSequence[str]:
        allocation = self.get_allocation(job_name)
        return allocation.locations if allocation is not None and (
                statuses is None or allocation.status in statuses) else []

    def get_service(self, job_name: str) -> Optional[str]:
        allocation = self.get_allocation(job_name)
        return allocation.target.service if allocation else None

    @abstractmethod
    async def notify_status(self,
                            job_name: str,
                            status: Status):
        ...

    @abstractmethod
    async def schedule(self,
                       job: Job,
                       target: Target,
                       hardware_requirement: Hardware) -> None:
        ...



from __future__ import annotations

from abc import ABC, abstractmethod
from asyncio import Lock
from enum import Enum
from typing import Any, MutableMapping, MutableSequence, Optional, TYPE_CHECKING, Type

from streamflow.core.config import SchemaEntity
from streamflow.core.context import StreamFlowContext

if TYPE_CHECKING:
    from streamflow.core.deployment import DeploymentConfig, Target
    from streamflow.core.workflow import Port, Step, Token, Workflow


class DatabaseLoadingContext(ABC):

    @abstractmethod
    def add_deployment(self,
                       persistent_id: int,
                       deployment: DeploymentConfig):
        ...

    @abstractmethod
    def add_port(self,
                 persistent_id: int,
                 port: Port):
        ...

    @abstractmethod
    def add_step(self,
                 persistent_id: int,
                 step: Step):
        ...

    @abstractmethod
    def add_target(self,
                   persistent_id: int,
                   target: Target):
        ...

    @abstractmethod
    def add_token(self,
                  persistent_id: int,
                  token: Token):
        ...

    @abstractmethod
    def add_workflow(self,
                     persistent_id: int,
                     workflow: Workflow):
        ...

    @abstractmethod
    async def load_deployment(self,
                              context: StreamFlowContext,
                              persistent_id: int):
        ...

    @abstractmethod
    async def load_port(self,
                        context: StreamFlowContext,
                        persistent_id: int):
        ...

    @abstractmethod
    async def load_step(self,
                        context: StreamFlowContext,
                        persistent_id: int):
        ...

    @abstractmethod
    async def load_target(self,
                          context: StreamFlowContext,
                          persistent_id: int):
        ...

    @abstractmethod
    async def load_token(self,
                         context: StreamFlowContext,
                         persistent_id: int):
        ...

    @abstractmethod
    async def load_workflow(self,
                            context: StreamFlowContext,
                            persistent_id: int):
        ...


class PersistableEntity(object):

    def __init__(self):
        self.persistent_id: Optional[int] = None
        self.persistence_lock: Lock = Lock()

    @classmethod
    @abstractmethod
    async def load(cls,
                   context: StreamFlowContext,
                   persistent_id: int,
                   loading_context: DatabaseLoadingContext) -> PersistableEntity:
        ...

    @abstractmethod
    async def save(self, context: StreamFlowContext) -> None:
        ...


class DependencyType(Enum):
    INPUT = 0
    OUTPUT = 1


class Database(SchemaEntity):

    def __init__(self, context: StreamFlowContext):
        self.context: StreamFlowContext = context

    @abstractmethod
    async def add_command(self,
                          step_id: int,
                          cmd: str) -> int:
        ...

    @abstractmethod
    async def add_dependency(self,
                             step: int,
                             port: int,
                             type: DependencyType, name: str) -> None:
        ...

    @abstractmethod
    async def add_deployment(self,
                             name: str,
                             type: str,
                             config: str,
                             external: bool,
                             lazy: bool,
                             workdir: Optional[str]) -> int:
        ...

    @abstractmethod
    async def add_port(self,
                       name: str,
                       workflow_id: int,
                       type: Type[Port],
                       params: str) -> int:
        ...

    @abstractmethod
    async def add_provenance(self,
                             inputs: MutableSequence[int],
                             token: int):
        ...

    @abstractmethod
    async def add_step(self,
                       name: str,
                       workflow_id: int,
                       status: int,
                       type: Type[Step],
                       params: str) -> int:
        ...

    @abstractmethod
    async def add_target(self,
                         deployment: int,
                         type: Type[Target],
                         params: str,
                         locations: int = 1,
                         service: Optional[str] = None,
                         workdir: Optional[str] = None) -> int:
        ...

    @abstractmethod
    async def add_token(self,
                        tag: str,
                        type: Type[Token],
                        value: Any,
                        port: Optional[int] = None) -> int:
        ...

    @abstractmethod
    async def add_workflow(self,
                           name: str,
                           params: str,
                           status: int,
                           type: str) -> int:
        ...

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    async def get_deployment(self, deplyoment_id: int) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    async def get_input_ports(self, step_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        ...

    @abstractmethod
    async def get_output_ports(self, step_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        ...

    @abstractmethod
    async def get_port(self, port_id: int) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    async def get_port_tokens(self, port_id: int) -> MutableSequence[int]:
        ...

    @abstractmethod
    async def get_report(self, workflow: str) -> MutableSequence[MutableMapping[str, Any]]:
        ...

    @abstractmethod
    async def get_step(self, step_id: int) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    async def get_target(self, target_id: int) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    async def get_token(self, token_id: int) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    async def get_workflow(self, workflow_id: int) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    async def get_workflow_ports(self, workflow_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        ...

    @abstractmethod
    async def get_workflow_steps(self, workflow_id: int) -> MutableSequence[MutableMapping[str, Any]]:
        ...

    @abstractmethod
    async def list_workflows(self) -> MutableSequence[MutableMapping[str, Any]]:
        ...

    @abstractmethod
    async def update_command(self,
                             command_id: int,
                             updates: MutableMapping[str, Any]) -> int:
        ...

    @abstractmethod
    async def update_deployment(self,
                                deployment_id: int,
                                updates: MutableMapping[str, Any]) -> int:
        ...

    @abstractmethod
    async def update_port(self,
                          port_id: int,
                          updates: MutableMapping[str, Any]) -> int:
        ...

    @abstractmethod
    async def update_step(self,
                          step_id: int,
                          updates: MutableMapping[str, Any]) -> int:
        ...

    @abstractmethod
    async def update_target(self,
                            target_id: str,
                            updates: MutableMapping[str, Any]) -> int:
        ...

    @abstractmethod
    async def update_workflow(self,
                              workflow_id: int,
                              updates: MutableMapping[str, Any]) -> int:
        ...



from abc import ABC

from streamflow.core.exception import WorkflowDefinitionException
from streamflow.core.workflow import Port
from streamflow.workflow.step import Transformer


class ManyToOneTransformer(Transformer, ABC):

    def add_output_port(self, name: str, port: Port) -> None:
        if not self.output_ports or port.name in self.output_ports:
            super().add_output_port(name, port)
        else:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))

    def get_output_name(self):
        return next(iter(self.output_ports))

    async def run(self):
        if len(self.output_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))
        await super().run()


class OneToOneTransformer(ManyToOneTransformer, ABC):

    def add_input_port(self, name: str, port: Port) -> None:
        if not self.input_ports:
            super().add_input_port(name, port)
        else:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))

    async def run(self):
        if len(self.input_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))
        if len(self.output_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))
        await super().run()



from __future__ import annotations

from typing import Any, AsyncIterable, MutableMapping, MutableSequence

from streamflow.core.context import StreamFlowContext
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.utils import get_tag
from streamflow.core.workflow import Token, Workflow
from streamflow.workflow.combinator import DotProductCombinator
from streamflow.workflow.token import IterationTerminationToken, ListToken


def _flatten_token_list(outputs: MutableSequence[Token]):
    flattened_list = []
    for token in sorted(outputs, key=lambda t: int(t.tag.split('.')[-1])):
        if isinstance(token, ListToken):
            flattened_list.extend(_flatten_token_list(token.value))
        else:
            flattened_list.append(token)
    return flattened_list


class ListMergeCombinator(DotProductCombinator):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 input_names: MutableSequence[str],
                 output_name: str,
                 flatten: bool = False):
        super().__init__(name, workflow)
        self.flatten: bool = flatten
        self.input_names: MutableSequence[str] = input_names
        self.output_name: str = output_name
        self.token_values: MutableMapping[str, MutableMapping[str, Any]] = {}

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> ListMergeCombinator:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            input_names=row['input_names'],
            output_name=row['output_name'],
            flatten=row['flatten'])

    async def _save_additional_params(self, context: StreamFlowContext):
        return {**await super()._save_additional_params(context),
                **{'input_names': self.input_names,
                   'output_name': self.output_name,
                   'flatten': self.flatten}}

    async def combine(self,
                      port_name: str,
                      token: Token) -> AsyncIterable[MutableMapping[str, Token]]:
        if not isinstance(token, IterationTerminationToken):
            async for schema in super().combine(port_name, token):
                outputs = [schema[name] for name in self.input_names]
                # Flatten if needed
                if self.flatten:
                    outputs = _flatten_token_list(outputs)
                yield {self.output_name: ListToken(
                    value=outputs,
                    tag=get_tag(outputs))}



from __future__ import annotations

import asyncio
import json
from abc import ABC, abstractmethod
from typing import Any, MutableMapping, MutableSequence

from streamflow.core.context import StreamFlowContext
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.workflow import Token


class IterationTerminationToken(Token):

    def __init__(self, tag: str):
        super().__init__(None, tag)

    def get_weight(self, context: StreamFlowContext):
        return 0

    def update(self, value: Any) -> Token:
        raise NotImplementedError()

    def retag(self, tag: str) -> Token:
        raise NotImplementedError()


class FileToken(Token, ABC):

    @abstractmethod
    async def get_paths(self, context: StreamFlowContext) -> MutableSequence[str]:
        ...


class JobToken(Token):

    async def _save_value(self, context: StreamFlowContext):
        return self.value.name


class ListToken(Token):

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> Token:
        value = json.loads(row['value'])
        return cls(
            tag=row['tag'],
            value=await asyncio.gather(*(asyncio.create_task(loading_context.load_token(context, t)) for t in value)))

    async def _save_value(self, context: StreamFlowContext):
        await asyncio.gather(*(asyncio.create_task(t.save(context)) for t in self.value))
        return [t.persistent_id for t in self.value]

    async def get_weight(self, context: StreamFlowContext):
        return sum(await asyncio.gather(*(asyncio.create_task(t.get_weight(context)) for t in self.value)))

    async def is_available(self, context: StreamFlowContext):
        return all(asyncio.gather(*(asyncio.create_task(t.is_available(context)) for t in self.value)))


class ObjectToken(Token):

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> Token:
        value = json.loads(row['value'])
        return cls(
            tag=row['tag'],
            value={k: v for k, v in zip(
                value.keys(),
                await asyncio.gather(*(asyncio.create_task(loading_context.load_token(context, v))
                                       for v in value.values())))})

    async def _save_value(self, context: StreamFlowContext):
        await asyncio.gather(*(asyncio.create_task(t.save(context)) for t in self.value.values()))
        return {k: t.persistent_id for k, t in self.value.items()}

    async def get_weight(self, context: StreamFlowContext):
        return sum(await asyncio.gather(*(asyncio.create_task(t.get_weight(context)) for t in self.value.values())))

    async def is_available(self, context: StreamFlowContext):
        return all(asyncio.gather(*(asyncio.create_task(t.is_available(context)) for t in self.value.values())))


class TerminationToken(Token):

    def __init__(self):
        super().__init__(None)

    def get_weight(self, context: StreamFlowContext):
        return 0

    def update(self, value: Any) -> Token:
        raise NotImplementedError()

    def retag(self, tag: str) -> Token:
        raise NotImplementedError()



from __future__ import annotations

import asyncio
import os
from asyncio import Condition
from typing import MutableSequence, Optional, TYPE_CHECKING

import pkg_resources

from streamflow.core.config import Config
from streamflow.core.data import LOCAL_LOCATION
from streamflow.core.deployment import Target
from streamflow.core.scheduling import Hardware, JobAllocation, LocationAllocation, Policy, Scheduler
from streamflow.core.workflow import Job, Status
from streamflow.log_handler import logger
from streamflow.scheduling.policy import policy_classes

if TYPE_CHECKING:
    from streamflow.core.context import StreamFlowContext
    from streamflow.core.scheduling import Location
    from typing import MutableMapping


class DefaultScheduler(Scheduler):

    def __init__(self,
                 context: StreamFlowContext,
                 retry_delay: Optional[int] = None) -> None:
        super().__init__(context)
        self.allocation_groups: MutableMapping[str, MutableSequence[Job]] = {}
        self.policy_map: MutableMapping[str, Policy] = {}
        self.retry_interval: Optional[int] = retry_delay
        self.scheduling_groups: MutableMapping[str, MutableSequence[str]] = {}
        self.wait_queues: MutableMapping[str, Condition] = {}

    def _allocate_job(self,
                      job: Job,
                      hardware: Hardware,
                      selected_locations: MutableSequence[str],
                      target: Target):
        if len(selected_locations) == 1:
            logger.debug(
                "Job {name} allocated {location}".format(
                    name=job.name,
                    location=("locally" if selected_locations[0] == LOCAL_LOCATION else
                              "on location {loc}".format(loc=selected_locations[0]))))
        else:
            logger.debug(
                "Job {name} allocated on locations {locations}".format(
                    name=job.name,
                    locations=', '.join(selected_locations)))
        self.job_allocations[job.name] = JobAllocation(
            job=job.name,
            target=target,
            locations=selected_locations,
            status=Status.FIREABLE,
            hardware=hardware or Hardware())
        for selected_location in selected_locations:
            if selected_location not in self.location_allocations:
                self.location_allocations[selected_location] = LocationAllocation(
                    name=selected_location,
                    deployment=target.deployment.name)
            self.location_allocations[selected_location].jobs.append(job.name)

    def _deallocate_job(self, job: str):
        job_allocation = self.job_allocations.pop(job)
        for selected_location in job_allocation.locations:
            self.location_allocations[selected_location].jobs.remove(job)
        if len(job_allocation.locations) == 1:
            logger.info(
                "Job {name} deallocated {location}".format(
                    name=job,
                    location=("from local location" if job_allocation.locations[0] == LOCAL_LOCATION else
                              "from location {loc}".format(loc=job_allocation.locations[0]))))
        else:
            logger.info(
                "Job {name} deallocated from locations {locations}".format(
                    name=job,
                    locations=', '.join(job_allocation.locations)))

    async def _get_locations(self,
                             job: Job,
                             deployment: str,
                             hardware_requirement: Hardware,
                             locations: int,
                             scheduling_policy: Policy,
                             available_locations: MutableMapping[str, Location]) -> Optional[MutableSequence[str]]:
        selected_locations = []
        for i in range(locations):
            selected_location = await scheduling_policy.get_location(
                context=self.context,
                job=job,
                deployment=deployment,
                hardware_requirement=hardware_requirement,
                available_locations=available_locations,
                jobs=self.job_allocations,
                locations=self.location_allocations)
            if selected_location is not None:
                selected_locations.append(selected_location)
                available_locations.pop(selected_location)
            else:
                return None
        return selected_locations

    def _get_policy(self, policy_config: Config):
        if policy_config.name not in self.policy_map:
            self.policy_map[policy_config.name] = policy_classes[policy_config.type](**policy_config.config)
        return self.policy_map[policy_config.name]

    async def close(self):
        pass

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'scheduler.json'))

    async def notify_status(self, job_name: str, status: Status) -> None:
        connector = self.get_connector(job_name)
        if connector:
            if connector.deployment_name in self.wait_queues:
                async with self.wait_queues[connector.deployment_name]:
                    if job_name in self.job_allocations:
                        if status != self.job_allocations[job_name].status:
                            self.job_allocations[job_name].status = status
                            logger.debug(
                                "Job {name} changed status to {status}".format(name=job_name, status=status.name))
                        if status in [Status.COMPLETED, Status.FAILED]:
                            self.wait_queues[connector.deployment_name].notify_all()

    async def schedule(self,
                       job: Job,
                       target: Target,
                       hardware_requirement: Hardware):
        deployment = target.deployment.name
        if deployment not in self.wait_queues:
            self.wait_queues[deployment] = Condition()
        async with self.wait_queues[deployment]:
            while True:
                connector = self.context.deployment_manager.get_connector(target.deployment.name)
                logger.debug("Retreiving available locations for job {}".format(job.name))
                available_locations = dict(await connector.get_available_locations(
                    service=target.service,
                    input_directory=job.input_directory,
                    output_directory=job.output_directory,
                    tmp_directory=job.tmp_directory))
                logger.debug("Available locations for job {} are {}".format(
                    job.name, {k: l.hostname for k, l in available_locations.items()}))
                if available_locations:
                    if target.scheduling_group is not None:
                        if target.scheduling_group not in self.allocation_groups:
                            self.allocation_groups[target.scheduling_group] = []
                        self.allocation_groups[target.scheduling_group].append(job)
                        group_size = len(self.scheduling_groups[target.scheduling_group])
                        if len(self.allocation_groups.get(target.scheduling_group, [])) == group_size:
                            allocated_jobs = []
                            for j in self.allocation_groups[target.scheduling_group]:
                                selected_locations = await self._get_locations(
                                    job=job,
                                    deployment=target.deployment.name,
                                    hardware_requirement=hardware_requirement,
                                    locations=target.locations,
                                    scheduling_policy=self._get_policy(target.scheduling_policy),
                                    available_locations=available_locations)
                                if selected_locations is None:
                                    break
                                self._allocate_job(
                                    job=j,
                                    hardware=hardware_requirement,
                                    selected_locations=selected_locations,
                                    target=target)
                                allocated_jobs.append(j)
                            if len(allocated_jobs) < group_size:
                                for j in allocated_jobs:
                                    self._deallocate_job(j.name)
                            else:
                                return
                    else:
                        selected_locations = await self._get_locations(
                            job=job,
                            deployment=target.deployment.name,
                            hardware_requirement=hardware_requirement,
                            locations=target.locations,
                            scheduling_policy=self._get_policy(target.scheduling_policy),
                            available_locations=available_locations)
                        if selected_locations is not None:
                            self._allocate_job(
                                job=job,
                                hardware=hardware_requirement,
                                selected_locations=selected_locations,
                                target=target)
                            return
                try:
                    await asyncio.wait_for(self.wait_queues[deployment].wait(), timeout=self.retry_interval)
                except asyncio.TimeoutError:
                    pass



from __future__ import annotations

import asyncio
import contextlib
import os
import posixpath
import stat
import tarfile
import tempfile
from asyncio import Lock, Semaphore
from asyncio.subprocess import STDOUT
from pathlib import PurePosixPath
from typing import Any, MutableMapping, MutableSequence, Optional, Tuple, Union

import asyncssh
import pkg_resources
from asyncssh import SSHClientConnection
from asyncssh.process import SSHProcess
from cachetools import Cache, LRUCache
from jinja2 import Template

from streamflow.core import utils
from streamflow.core.asyncache import cachedmethod
from streamflow.core.context import StreamFlowContext
from streamflow.core.data import StreamWrapperContext
from streamflow.core.deployment import Connector
from streamflow.core.exception import WorkflowExecutionException
from streamflow.core.scheduling import Hardware, Location
from streamflow.data import aiotarstream
from streamflow.data.stream import StreamReaderWrapper, StreamWriterWrapper
from streamflow.deployment.connector.base import BaseConnector
from streamflow.log_handler import logger


async def _get_disk_usage(ssh_client: SSHClientConnection, directory: str) -> float:
    if directory:
        result = await ssh_client.run("df {} | tail -n 1 | awk '{{print $2}}'".format(directory), stderr=STDOUT)
        if result.returncode == 0:
            return float(result.stdout.strip()) / 2 ** 10
        else:
            raise WorkflowExecutionException(result.returncode)
    else:
        return float('inf')


def _parse_hostname(hostname):
    if ':' in hostname:
        hostname, port = hostname.split(':')
        port = int(port)
    else:
        port = 22
    return hostname, port


class SSHContext(object):

    def __init__(self,
                 streamflow_config_dir: str,
                 config: SSHConfig,
                 max_concurrent_sessions: int):
        self._streamflow_config_dir: str = streamflow_config_dir
        self._config: SSHConfig = config
        self._ssh_connection: Optional[SSHClientConnection] = None
        self._connection_lock: Lock = Lock()
        self._sem: Semaphore = Semaphore(max_concurrent_sessions)

    async def __aenter__(self):
        async with self._connection_lock:
            if self._ssh_connection is None:
                self._ssh_connection = await self._get_connection(self._config)
        await self._sem.acquire()
        return self._ssh_connection

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._sem.release()

    async def _get_connection(self,
                              config: SSHConfig):
        if config is None:
            return None
        (hostname, port) = _parse_hostname(config.hostname)
        passphrase = (self._get_param_from_file(config.ssh_key_passphrase_file)
                      if config.ssh_key_passphrase_file else None)
        password = self._get_param_from_file(config.password_file) if config.password_file else None
        return await asyncssh.connect(
            client_keys=config.client_keys,
            compression_algs=None,
            encryption_algs=[
                "aes128-gcm@openssh.com",
                "aes256-ctr",
                "aes192-ctr",
                "aes128-ctr"
            ],
            known_hosts=() if config.check_host_key else None,
            host=hostname,
            passphrase=passphrase,
            password=password,
            port=port,
            tunnel=await self._get_connection(config.tunnel),
            username=config.username)

    def _get_param_from_file(self,
                             file_path: str):
        if not os.path.isabs(file_path):
            file_path = os.path.join(self._streamflow_config_dir, file_path)
        with open(file_path) as f:
            return f.read().strip()


class SSHStreamWrapperContext(StreamWrapperContext):

    def __init__(self,
                 src: str,
                 ssh_context: SSHContext):
        super().__init__()
        self.src: str = src
        self.ssh_context: SSHContext = ssh_context
        self.ssh_process: Optional[SSHProcess] = None
        self.stream: Optional[StreamReaderWrapper] = None

    async def __aenter__(self):
        ssh_client = await self.ssh_context.__aenter__()
        dirname, basename = posixpath.split(self.src)
        self.ssh_process = await (await ssh_client.create_process(
            "tar chf - -C {} {}".format(dirname, basename),
            stdin=asyncio.subprocess.DEVNULL,
            encoding=None)).__aenter__()
        self.stream = StreamReaderWrapper(self.ssh_process.stdout)
        return self.stream

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.stream:
            await self.stream.close()
        if self.ssh_process:
            await self.ssh_process.__aexit__(exc_type, exc_val, exc_tb)
        await self.ssh_context.__aexit__(exc_type, exc_val, exc_tb)


class SSHConfig(object):

    def __init__(self,
                 check_host_key: bool,
                 client_keys: MutableSequence[str],
                 hostname: str,
                 password_file: Optional[str],
                 ssh_key_passphrase_file: Optional[str],
                 tunnel: Optional[SSHConfig],
                 username: str):
        self.check_host_key: bool = check_host_key
        self.client_keys: MutableSequence[str] = client_keys
        self.hostname: str = hostname
        self.password_file: Optional[str] = password_file
        self.ssh_key_passphrase_file: Optional[str] = ssh_key_passphrase_file
        self.tunnel: Optional[SSHConfig] = tunnel
        self.username: str = username


class SSHConnector(BaseConnector):

    @staticmethod
    def _get_command(location: str,
                     command: MutableSequence[str],
                     environment: MutableMapping[str, str] = None,
                     workdir: Optional[str] = None,
                     stdin: Optional[Union[int, str]] = None,
                     stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                     stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                     job_name: Optional[str] = None):
        command = utils.create_command(
            command=command,
            environment=environment,
            workdir=workdir,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr)
        logger.debug("Executing command {command} on {location} {job}".format(
            command=command,
            location=location,
            job="for job {job}".format(job=job_name) if job_name else ""))
        return utils.encode_command(command)

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 nodes: MutableSequence[Any],
                 username: str,
                 checkHostKey: bool = True,
                 dataTransferConnection: Optional[Union[str, MutableMapping[str, Any]]] = None,
                 file: Optional[str] = None,
                 maxConcurrentSessions: int = 10,
                 passwordFile: Optional[str] = None,
                 sharedPaths: Optional[MutableSequence[str]] = None,
                 sshKey: Optional[str] = None,
                 sshKeyPassphraseFile: Optional[str] = None,
                 tunnel: Optional[MutableMapping[str, Any]] = None,
                 transferBufferSize: int = 2 ** 16) -> None:
        super().__init__(
            deployment_name=deployment_name,
            context=context,
            transferBufferSize=transferBufferSize)
        if file is not None:
            with open(os.path.join(context.config_dir, file)) as f:
                self.template: Optional[Template] = Template(f.read())
        else:
            self.template: Optional[Template] = None
        self.checkHostKey: bool = checkHostKey
        self.passwordFile: Optional[str] = passwordFile
        self.maxConcurrentSessions: int = maxConcurrentSessions
        self.sharedPaths: MutableSequence[str] = sharedPaths or []
        self.sshKey: Optional[str] = sshKey
        self.sshKeyPassphraseFile: Optional[str] = sshKeyPassphraseFile
        self.ssh_contexts: MutableMapping[str, SSHContext] = {}
        self.data_transfer_contexts: MutableMapping[str, SSHContext] = {}
        self.username: str = username
        self.tunnel: Optional[SSHConfig] = self._get_config(tunnel)
        self.dataTransferConfig: Optional[SSHConfig] = self._get_config(
            dataTransferConnection)
        self.nodes: MutableMapping[str, SSHConfig] = {n.hostname: n for n in [self._get_config(n) for n in nodes]}
        self.hardwareCache: Cache = LRUCache(maxsize=len(self.nodes))

    async def _build_helper_file(self,
                                 command: str,
                                 location: str,
                                 environment: MutableMapping[str, str] = None,
                                 workdir: str = None) -> str:
        helper_file = tempfile.mktemp()
        with open(helper_file, mode='w') as f:
            f.write(self.template.render(
                streamflow_command="sh -c '{command}'".format(command=command),
                streamflow_environment=environment,
                streamflow_workdir=workdir))
        os.chmod(helper_file, os.stat(helper_file).st_mode | stat.S_IEXEC)
        remote_path = posixpath.join(workdir or '/tmp', os.path.basename(helper_file))
        await self._copy_local_to_remote(
            src=helper_file,
            dst=remote_path,
            locations=[location])
        return remote_path

    async def _copy_local_to_remote_single(self,
                                           src: str,
                                           dst: str,
                                           location: str,
                                           read_only: bool = False):
        async with self._get_data_transfer_client(location) as ssh_client:
            async with ssh_client.create_process(
                    "tar xf - -C /",
                    stderr=asyncio.subprocess.DEVNULL,
                    stdout=asyncio.subprocess.DEVNULL,
                    encoding=None) as proc:
                try:
                    async with aiotarstream.open(
                            stream=StreamWriterWrapper(proc.stdin),
                            format=tarfile.GNU_FORMAT,
                            mode='w',
                            dereference=True,
                            copybufsize=self.transferBufferSize) as tar:
                        await tar.add(src, arcname=dst)
                except tarfile.TarError as e:
                    raise WorkflowExecutionException("Error copying {} to {} on location {}: {}".format(
                        src, dst, location, str(e))) from e

    async def _copy_remote_to_local(self,
                                    src: str,
                                    dst: str,
                                    location: str,
                                    read_only: bool = False) -> None:
        async with self._get_data_transfer_client(location) as ssh_client:
            async with ssh_client.create_process(
                    "tar chf - -C / " + posixpath.relpath(src, '/'),
                    stdin=asyncio.subprocess.DEVNULL,
                    encoding=None) as proc:
                try:
                    async with aiotarstream.open(
                            stream=StreamReaderWrapper(proc.stdout),
                            mode='r',
                            copybufsize=self.transferBufferSize) as tar:
                        await utils.extract_tar_stream(tar, src, dst, self.transferBufferSize)
                except tarfile.TarError as e:
                    raise WorkflowExecutionException("Error copying {} from location {} to {}: {}".format(
                        src, location, dst, str(e))) from e

    async def _copy_remote_to_remote(self,
                                     src: str,
                                     dst: str,
                                     locations: MutableSequence[str],
                                     source_location: str,
                                     source_connector: Optional[Connector] = None,
                                     read_only: bool = False) -> None:
        source_connector = source_connector or self
        if source_connector == self and source_location in locations:
            if src != dst:
                command = ['/bin/cp', "-rf", src, dst]
                await self.run(source_location, command)
                locations.remove(source_location)
        write_command = " ".join(await utils.get_remote_to_remote_write_command(
            src_connector=source_connector,
            src_location=source_location,
            src=src,
            dst_connector=self,
            dst_locations=locations,
            dst=dst))
        if locations:
            async with source_connector._get_stream_reader(source_location, src) as reader:
                async with contextlib.AsyncExitStack() as exit_stack:
                    # Open a target StreamWriter for each location
                    writer_clients = await asyncio.gather(*(asyncio.create_task(
                        exit_stack.enter_async_context(self._get_data_transfer_client(location))
                    ) for location in locations))
                    async with contextlib.AsyncExitStack() as writers_stack:
                        writers = await asyncio.gather(*(asyncio.create_task(
                            writers_stack.enter_async_context(
                                client.create_process(
                                    write_command,
                                    stderr=asyncio.subprocess.DEVNULL,
                                    stdout=asyncio.subprocess.DEVNULL,
                                    encoding=None)
                            )) for client in writer_clients))
                        # Multiplex the reader output to all the writers
                        while content := await reader.read(source_connector.transferBufferSize):
                            for writer in writers:
                                writer.stdin.write(content)
                            await asyncio.gather(*(asyncio.create_task(writer.stdin.drain()) for writer in writers))

    def _get_config(self,
                    node: Union[str, MutableMapping[str, Any]]):
        if node is None:
            return None
        elif isinstance(node, str):
            node = {'hostname': node}
        ssh_key = node['sshKey'] if 'sshKey' in node else self.sshKey
        return SSHConfig(
            hostname=node['hostname'],
            username=node['username'] if 'username' in node else self.username,
            check_host_key=node['checkHostKey'] if 'checkHostKey' in node else self.checkHostKey,
            client_keys=[ssh_key] if ssh_key is not None else [],
            password_file=node['passwordFile'] if 'passwordFile' in node else self.passwordFile,
            ssh_key_passphrase_file=(node['sshKeyPassphraseFile'] if 'sshKeyPassphraseFile' in node else
                                     self.sshKeyPassphraseFile),
            tunnel=(self._get_config(node['tunnel']) if 'tunnel' in node else
                    self.tunnel if hasattr(self, 'tunnel') else
                    None))

    def _get_data_transfer_client(self, location: str):
        if self.dataTransferConfig:
            if location not in self.data_transfer_contexts:
                self.data_transfer_contexts[location] = SSHContext(
                    streamflow_config_dir=self.context.config_dir,
                    config=self.dataTransferConfig,
                    max_concurrent_sessions=self.maxConcurrentSessions)
            return self.data_transfer_contexts[location]
        else:
            return self._get_ssh_client(location)

    async def _get_existing_parent(self, location: str, directory: str):
        if directory is None:
            return None
        command_template = "test -e \"{path}\""
        async with self._get_ssh_client(location) as ssh_client:
            while True:
                result = await ssh_client.run(command_template.format(path=directory), stderr=STDOUT)
                if result.returncode == 0:
                    return directory
                elif result.returncode == 1:
                    directory = PurePosixPath(directory).parent
                else:
                    raise WorkflowExecutionException(result.stdout.strip())

    @cachedmethod(lambda self: self.hardwareCache)
    async def _get_location_hardware(self,
                                     location: str,
                                     input_directory: str,
                                     output_directory: str,
                                     tmp_directory: str) -> Hardware:
        try:
            async with self._get_ssh_client(location) as ssh_client:
                cores, memory, input_directory, output_directory, tmp_directory = await asyncio.gather(
                    asyncio.create_task(ssh_client.run("nproc", stderr=STDOUT)),
                    asyncio.create_task(ssh_client.run("free | grep Mem | awk '{print $2}'", stderr=STDOUT)),
                    asyncio.create_task(_get_disk_usage(ssh_client, input_directory)),
                    asyncio.create_task(_get_disk_usage(ssh_client, output_directory)),
                    asyncio.create_task(_get_disk_usage(ssh_client, tmp_directory)))
                if (cores.returncode == 0 and
                        memory.returncode == 0):
                    return Hardware(
                        cores=float(cores.stdout.strip()),
                        memory=float(memory.stdout.strip()) / 2 ** 10,
                        input_directory=input_directory,
                        output_directory=output_directory,
                        tmp_directory=tmp_directory)
                else:
                    raise WorkflowExecutionException(
                        "Impossible to retrieve locations for {location}".format(location=location))
        except WorkflowExecutionException:
            raise WorkflowExecutionException(
                "Impossible to retrieve locations for {location}".format(location=location))

    def _get_run_command(self,
                         command: str,
                         location: str,
                         interactive: bool = False):
        return "".join([
            "ssh ",
            "{location} ",
            "{command}"
        ]).format(
            location=location,
            command=command)

    def _get_ssh_client(self, location: str):
        if location not in self.ssh_contexts:
            self.ssh_contexts[location] = SSHContext(
                streamflow_config_dir=self.context.config_dir,
                config=self.nodes[location],
                max_concurrent_sessions=self.maxConcurrentSessions)
        return self.ssh_contexts[location]

    def _get_stream_reader(self,
                           location: str,
                           src: str) -> StreamWrapperContext:
        return SSHStreamWrapperContext(
            src=src,
            ssh_context=self._get_data_transfer_client(location))

    async def deploy(self, external: bool) -> None:
        pass

    async def get_available_locations(self,
                                      service: str,
                                      input_directory: Optional[str] = None,
                                      output_directory: Optional[str] = None,
                                      tmp_directory: Optional[str] = None) -> MutableMapping[str, Location]:
        locations = {}
        for location_obj in self.nodes.values():
            inpdir, outdir, tmpdir = await asyncio.gather(
                asyncio.create_task(self._get_existing_parent(location_obj.hostname, input_directory)),
                asyncio.create_task(self._get_existing_parent(location_obj.hostname, output_directory)),
                asyncio.create_task(self._get_existing_parent(location_obj.hostname, tmp_directory)))
            hardware = await self._get_location_hardware(
                location=location_obj.hostname,
                input_directory=inpdir,
                output_directory=outdir,
                tmp_directory=tmpdir)
            locations[location_obj.hostname] = Location(
                name=location_obj.hostname,
                hostname=location_obj.hostname,
                hardware=hardware)
        return locations

    @classmethod
    def get_schema(cls) -> str:
        return pkg_resources.resource_filename(
            __name__, os.path.join('schemas', 'ssh.json'))

    async def run(self,
                  location: str,
                  command: MutableSequence[str],
                  environment: MutableMapping[str, str] = None,
                  workdir: Optional[str] = None,
                  stdin: Optional[Union[int, str]] = None,
                  stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                  stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                  capture_output: bool = False,
                  job_name: Optional[str] = None) -> Optional[Tuple[Optional[Any], int]]:
        command = self._get_command(
            location=location,
            command=command,
            environment=environment,
            workdir=workdir,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            job_name=job_name)
        if job_name is not None and self.template is not None:
            helper_file = await self._build_helper_file(command, location, environment, workdir)
            async with self._get_ssh_client(location) as ssh_client:
                result = await ssh_client.run(helper_file, stderr=STDOUT)
        else:
            async with self._get_ssh_client(location) as ssh_client:
                result = await ssh_client.run("sh -c '{command}'".format(command=command), stderr=STDOUT)
        if capture_output:
            return result.stdout.strip(), result.returncode

    async def undeploy(self, external: bool) -> None:
        for ssh_context in self.ssh_contexts.values():
            async with ssh_context as ssh_client:
                ssh_client.close()
        self.ssh_contexts = {}
        for ssh_context in self.data_transfer_contexts.values():
            async with ssh_context as ssh_client:
                ssh_client.close()
        self.data_transfer_contexts = {}



from __future__ import annotations

import asyncio
from concurrent.futures import ProcessPoolExecutor
from typing import TYPE_CHECKING

from streamflow.core.recovery import CheckpointManager, FailureManager
from streamflow.log_handler import logger

if TYPE_CHECKING:
    from streamflow.core.data import DataManager
    from streamflow.core.deployment import DeploymentManager
    from streamflow.core.persistence import Database
    from streamflow.core.scheduling import Scheduler
    from typing import Optional


class StreamFlowContext(object):

    def __init__(self, streamflow_config_dir: str):
        self.config_dir = streamflow_config_dir
        self.checkpoint_manager: Optional[CheckpointManager] = None
        self.database: Optional[Database] = None
        self.data_manager: Optional[DataManager] = None
        self.deployment_manager: Optional[DeploymentManager] = None
        self.failure_manager: Optional[FailureManager] = None
        self.process_executor: ProcessPoolExecutor = ProcessPoolExecutor()
        self.scheduler: Optional[Scheduler] = None

    async def close(self):
        try:
            await asyncio.gather(
                self.checkpoint_manager.close(),
                self.data_manager.close(),
                self.deployment_manager.close(),
                self.failure_manager.close(),
                self.scheduler.close())
        except BaseException as e:
            logger.exception(e)
        finally:
            await self.database.close()



from typing import MutableMapping, Optional

from streamflow.core.workflow import Job, Step, Token


class JobVersion(object):
    __slots__ = ('job', 'outputs', 'step', 'version')

    def __init__(self,
                 job: Job,
                 outputs: Optional[MutableMapping[str, Token]],
                 step: Step,
                 version: int = 1):
        self.job: Job = job
        self.outputs: Optional[MutableMapping[str, Token]] = outputs
        self.step: Step = step
        self.version: int = version



from __future__ import annotations

import asyncio
import posixpath
import shlex
import tarfile
from abc import ABC, abstractmethod
from typing import MutableSequence, TYPE_CHECKING, Tuple

from streamflow.core import utils
from streamflow.core.context import StreamFlowContext
from streamflow.core.data import LOCAL_LOCATION, StreamWrapperContext
from streamflow.core.deployment import Connector, ConnectorCopyKind
from streamflow.core.exception import WorkflowExecutionException
from streamflow.data import aiotarstream
from streamflow.data.stream import StreamReaderWrapper, StreamWriterWrapper, SubprocessStreamReaderWrapperContext
from streamflow.log_handler import logger

if TYPE_CHECKING:
    from typing import Any, Optional, MutableMapping, Union


class BaseConnector(Connector, ABC):

    @staticmethod
    def get_option(name: str,
                   value: Any,
                   ) -> str:
        if len(name) > 1:
            name = "-{name} ".format(name=name)
        if isinstance(value, bool):
            return "-{name} ".format(name=name) if value else ""
        elif isinstance(value, str):
            return "-{name} \"{value}\" ".format(name=name, value=value)
        elif isinstance(value, MutableSequence):
            return "".join(["-{name} \"{value}\" ".format(name=name, value=item) for item in value])
        elif value is None:
            return ""
        else:
            raise TypeError("Unsupported value type")

    def __init__(self,
                 deployment_name: str,
                 context: StreamFlowContext,
                 transferBufferSize: int):
        super().__init__(deployment_name, context)
        self.transferBufferSize: int = transferBufferSize
        self.is_deployed: bool = False

    async def _copy_local_to_remote(self,
                                    src: str,
                                    dst: str,
                                    locations: MutableSequence[str],
                                    read_only: bool = False) -> None:
        await asyncio.gather(*(asyncio.create_task(
            self._copy_local_to_remote_single(
                src=src,
                dst=dst,
                location=location,
                read_only=read_only)
        ) for location in locations))

    async def _copy_local_to_remote_single(self,
                                           src: str,
                                           dst: str,
                                           location: str,
                                           read_only: bool = False) -> None:
        proc = await asyncio.create_subprocess_exec(
            *shlex.split(self._get_run_command(
                command="tar xf - -C /",
                location=location,
                interactive=True)),
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL)
        try:
            async with aiotarstream.open(
                    stream=StreamWriterWrapper(proc.stdin),
                    format=tarfile.GNU_FORMAT,
                    mode='w',
                    dereference=True,
                    copybufsize=self.transferBufferSize) as tar:
                await tar.add(src, arcname=dst)
        except tarfile.TarError as e:
            raise WorkflowExecutionException("Error copying {} to {} on location {}: {}".format(
                src, dst, location, str(e))) from e
        finally:
            proc.stdin.close()
            await proc.wait()

    async def _copy_remote_to_local(self,
                                    src: str,
                                    dst: str,
                                    location: str,
                                    read_only: bool = False) -> None:
        proc = await asyncio.create_subprocess_exec(
            *shlex.split(self._get_run_command(
                command="tar chf - -C / " + posixpath.relpath(src, '/'),
                location=location)),
            stdin=None,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        try:
            async with aiotarstream.open(
                    stream=StreamReaderWrapper(proc.stdout),
                    mode='r',
                    copybufsize=self.transferBufferSize) as tar:
                await utils.extract_tar_stream(tar, src, dst, self.transferBufferSize)
        except tarfile.TarError as e:
            raise WorkflowExecutionException("Error copying {} from location {} to {}: {}".format(
                src, location, dst, str(e))) from e
        finally:
            await proc.wait()

    async def _copy_remote_to_remote(self,
                                     src: str,
                                     dst: str,
                                     locations: MutableSequence[str],
                                     source_location: str,
                                     source_connector: Optional[str] = None,
                                     read_only: bool = False) -> None:
        source_connector = source_connector or self
        if source_connector == self and source_location in locations:
            if src != dst:
                command = ['/bin/cp', "-rf", src, dst]
                await self.run(source_location, command)
                locations.remove(source_location)
        if locations:
            # Get write command
            write_command = " ".join(await utils.get_remote_to_remote_write_command(
                src_connector=source_connector,
                src_location=source_location,
                src=src,
                dst_connector=self,
                dst_locations=locations,
                dst=dst))
            # Open source StreamReader
            async with source_connector._get_stream_reader(source_location, src) as reader:
                # Open a target StreamWriter for each location
                writers = await asyncio.gather(*(asyncio.create_task(asyncio.create_subprocess_exec(
                    *shlex.split(self._get_run_command(
                        command=write_command,
                        location=location,
                        interactive=True)),
                    stdin=asyncio.subprocess.PIPE,
                    stdout=asyncio.subprocess.DEVNULL,
                    stderr=asyncio.subprocess.DEVNULL)) for location in locations))
                try:
                    # Multiplex the reader output to all the writers
                    while content := await reader.read(source_connector.transferBufferSize):
                        for writer in writers:
                            writer.stdin.write(content)
                        await asyncio.gather(*(asyncio.create_task(writer.stdin.drain()) for writer in writers))
                finally:
                    # Close all writers
                    for writer in writers:
                        writer.stdin.close()
                    await asyncio.gather(*(asyncio.create_task(writer.wait()) for writer in writers))

    @abstractmethod
    def _get_run_command(self,
                         command: str,
                         location: str,
                         interactive: bool = False) -> str:
        ...

    def _get_stream_reader(self,
                           location: str,
                           src: str) -> StreamWrapperContext:
        dirname, basename = posixpath.split(src)
        return SubprocessStreamReaderWrapperContext(
            coro=asyncio.create_subprocess_exec(
                *shlex.split(self._get_run_command(
                    command="tar chf - -C {} {}".format(dirname, basename),
                    location=location)),
                stdin=None,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE))

    async def copy(self,
                   src: str,
                   dst: str,
                   locations: MutableSequence[str],
                   kind: ConnectorCopyKind,
                   source_connector: Optional[Connector] = None,
                   source_location: Optional[str] = None,
                   read_only: bool = False) -> None:
        if kind == ConnectorCopyKind.REMOTE_TO_REMOTE:
            if source_location is None:
                raise Exception("Source location is mandatory for remote to remote copy")
            if len(locations) > 1:
                logger.info("Copying {src} {source_location} to {dst} on locations:\n\t{locations}".format(
                    source_location=("on local file-system" if source_location == LOCAL_LOCATION else
                                     "on location {res}".format(res=source_location)),
                    src=src,
                    dst=dst,
                    locations='\n\t'.join(locations)
                ))
            else:
                logger.info("Copying {src} {source_location} to {dst} {location}".format(
                    source_location=("on local file-system" if source_location == LOCAL_LOCATION else
                                     "on location {res}".format(res=source_location)),
                    src=src,
                    dst=dst,
                    location=("on local file-system" if locations[0] == LOCAL_LOCATION else
                              "on location {res}".format(res=locations[0]))
                ))
            await self._copy_remote_to_remote(
                src=src,
                dst=dst,
                locations=locations,
                source_connector=source_connector,
                source_location=source_location,
                read_only=read_only)
        elif kind == ConnectorCopyKind.LOCAL_TO_REMOTE:
            if len(locations) > 1:
                logger.info("Copying {src} on local file-system to {dst} on locations:\n\t{locations}".format(
                    src=src,
                    dst=dst,
                    locations='\n\t'.join(locations)
                ))
            else:
                logger.info("Copying {src} on local file-system to {dst} on location {location}".format(
                    src=src,
                    dst=dst,
                    location=locations[0]
                ))
            await self._copy_local_to_remote(
                src=src,
                dst=dst,
                locations=locations,
                read_only=read_only)
        elif kind == ConnectorCopyKind.REMOTE_TO_LOCAL:
            if len(locations) > 1:
                raise Exception("Copy from multiple locations is not supported")
            logger.info("Copying {src} on location {location} to {dst} on local file-system".format(
                src=src,
                dst=dst,
                location=locations[0]
            ))
            await self._copy_remote_to_local(
                src=src,
                dst=dst,
                location=locations[0],
                read_only=read_only)
        else:
            raise NotImplementedError

    async def run(self,
                  location: str,
                  command: MutableSequence[str],
                  environment: MutableMapping[str, str] = None,
                  workdir: Optional[str] = None,
                  stdin: Optional[Union[int, str]] = None,
                  stdout: Union[int, str] = asyncio.subprocess.STDOUT,
                  stderr: Union[int, str] = asyncio.subprocess.STDOUT,
                  capture_output: bool = False,
                  job_name: Optional[str] = None) -> Optional[Tuple[Optional[Any], int]]:
        command = utils.create_command(
            command, environment, workdir, stdin, stdout, stderr)
        logger.debug("Executing command {command} on {location} {job}".format(
            command=command,
            location=location,
            job="for job {job}".format(job=job_name) if job_name else ""))
        command = utils.encode_command(command)
        run_command = self._get_run_command(command, location)
        proc = await asyncio.create_subprocess_exec(
            *shlex.split(run_command),
            stdin=None,
            stdout=asyncio.subprocess.PIPE if capture_output else asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.PIPE if capture_output else asyncio.subprocess.DEVNULL)
        if capture_output:
            stdout, _ = await proc.communicate()
            return stdout.decode().strip(), proc.returncode
        else:
            await proc.wait()



from __future__ import annotations

import asyncio
import json
import logging
import posixpath
from abc import ABC, abstractmethod
from asyncio import CancelledError, FIRST_COMPLETED, Task
from typing import Any, AsyncIterable, Iterable, MutableMapping, MutableSequence, Optional, Set, cast

from streamflow.core import utils
from streamflow.core.context import StreamFlowContext
from streamflow.core.deployment import Connector, DeploymentConfig, Target
from streamflow.core.exception import (
    FailureHandlingException, WorkflowDefinitionException, WorkflowExecutionException
)
from streamflow.core.persistence import DatabaseLoadingContext
from streamflow.core.scheduling import HardwareRequirement
from streamflow.core.utils import random_name
from streamflow.core.workflow import (
    Command, CommandOutput, CommandOutputProcessor, Job, Port, Status, Step, Token,
    Workflow
)
from streamflow.data import remotepath
from streamflow.log_handler import logger
from streamflow.workflow.port import ConnectorPort, JobPort
from streamflow.workflow.token import JobToken, ListToken, TerminationToken


def _get_step_status(statuses: MutableSequence[Status]):
    num_skipped = 0
    for status in statuses:
        if status == Status.FAILED:
            return Status.FAILED
        elif status == Status.CANCELLED:
            return Status.CANCELLED
        elif status == Status.SKIPPED:
            num_skipped += 1
    if num_skipped == len(statuses):
        return Status.SKIPPED
    else:
        return Status.COMPLETED


def _group_by_tag(inputs: MutableMapping[str, Token],
                  inputs_map: MutableMapping[str, MutableMapping[str, Token]]) -> None:
    for name, token in inputs.items():
        if token.tag not in inputs_map:
            inputs_map[token.tag] = {}
        inputs_map[token.tag][name] = token


class BaseStep(Step, ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__(name, workflow)
        self._log_level: int = logging.DEBUG

    async def _get_inputs(self,
                          input_ports: MutableMapping[str, Port]):
        inputs = {k: v for k, v in zip(input_ports.keys(), await asyncio.gather(*(
            asyncio.create_task(p.get(posixpath.join(self.name, port_name)))
            for port_name, p in input_ports.items())))}
        if utils.check_termination(inputs):
            logger.debug("Step {} received termination token".format(self.name))
        logger.debug("Step {} received inputs {}".format(self.name, [t.tag for t in inputs.values()]))
        return inputs

    async def _persist_token(self,
                             token: Token,
                             port: Port,
                             inputs: Iterable[Token]) -> Token:
        await token.save(self.workflow.context, port_id=port.persistent_id)
        if inputs:
            await self.workflow.context.database.add_provenance(
                inputs=[i.persistent_id for i in inputs],
                token=token.persistent_id)
        return token

    async def terminate(self, status: Status):
        if not self.terminated:
            self.terminated = True
            # If not explicitly cancelled, close input ports
            if status != Status.CANCELLED:
                for port_name, port in self.get_input_ports().items():
                    port.close(posixpath.join(self.name, port_name))
            # Add a TerminationToken to each output port
            for port in self.get_output_ports().values():
                port.put(TerminationToken())
            # Set termination status
            await self._set_status(status)
            logger.log(self._log_level, "Step {name} terminated with status {status}".format(
                name=self.name, status=status.name))


class Combinator(ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        self.name: str = name
        self.workflow: Workflow = workflow
        self.items: MutableSequence[str] = []
        self.combinators: MutableMapping[str, Combinator] = {}
        self.combinators_map: MutableMapping[str, str] = {}

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> Combinator:
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']))

    async def _save_additional_params(self, context: StreamFlowContext):
        return {
            "name": self.name,
            "combinators": {k: v for k, v in zip(
                self.combinators.keys(),
                await asyncio.gather(*(asyncio.create_task(comb.save(context))
                                       for comb in self.combinators.values())))},
            "combinators_map": self.combinators_map,
            "items": self.items,
            "workflow": self.workflow.persistent_id}

    def add_combinator(self, combinator: Combinator, items: Set[str]) -> None:
        self.combinators[combinator.name] = combinator
        self.items.append(combinator.name)
        self.combinators_map = {**self.combinators_map, **{p: combinator.name for p in items}}

    def add_item(self, item: str) -> None:
        self.items.append(item)

    def get_combinator(self, item: str) -> Optional[Combinator]:
        return self.combinators.get(self.combinators_map.get(item, ''))

    def get_items(self, recursive: bool = False) -> Set[str]:
        items = set(self.items)
        if recursive:
            for combinator in self.combinators.values():
                items.update(combinator.get_items(recursive))
        return items

    @abstractmethod
    async def combine(self,
                      port_name: str,
                      token: Token) -> AsyncIterable[MutableMapping[str, Token]]:
        ...

    @classmethod
    async def load(cls,
                   context: StreamFlowContext,
                   row: MutableMapping[str, Any],
                   loading_context: DatabaseLoadingContext) -> Combinator:
        type = cast(Combinator, utils.get_class_from_name(row['type']))
        combinator = await type._load(context, row['params'], loading_context)
        combinator.items = row['params']['items']
        combinator.combinators_map = row['params']['combinators_map']
        combinator.combinators = {}
        for k, c in row['params']['combinators'].items():
            combinator.combinators[k] = Combinator.load(context, c, loading_context)
        return combinator

    async def save(self, context: StreamFlowContext):
        return {'type': utils.get_class_fullname(type(self)),
                'params': await self._save_additional_params(context)}


class CombinatorStep(BaseStep):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 combinator: Combinator):
        super().__init__(name, workflow)
        self.combinator: Combinator = combinator

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> CombinatorStep:
        params = json.loads(row['params'])
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            combinator=await Combinator.load(context, params['combinator'], loading_context))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'combinator': await self.combinator.save(context)}}

    async def run(self):
        # Set default status to SKIPPED
        status = Status.SKIPPED
        if self.input_ports:
            input_tasks, terminated = [], []
            for port_name, port in self.get_input_ports().items():
                input_tasks.append(asyncio.create_task(
                    port.get(posixpath.join(self.name, port_name)), name=port_name))
            while input_tasks:
                # Wait for the next token
                finished, unfinished = await asyncio.wait(input_tasks, return_when=FIRST_COMPLETED)
                input_tasks = list(unfinished)
                for task in finished:
                    task_name = cast(Task, task).get_name()
                    token = task.result()
                    # If a TerminationToken is received, the corresponding port terminated its outputs
                    if utils.check_termination(token):
                        logger.debug("Step {} received termination token for port {}".format(
                            self.name, task_name))
                        terminated.append(task_name)
                    # Otherwise, build combination and set default status to COMPLETED
                    else:
                        logger.debug("Step {} received token {} on port {}".format(
                            self.name, token.tag, task_name))
                        status = Status.COMPLETED
                        async for schema in cast(AsyncIterable, self.combinator.combine(task_name, token)):
                            for port_name, token in schema.items():
                                self.get_output_port(port_name).put(await self._persist_token(
                                    token=token,
                                    port=self.get_output_port(port_name),
                                    inputs=schema.values()))
                    # Create a new task in place of the completed one if the port is not terminated
                    if task_name not in terminated:
                        input_tasks.append(asyncio.create_task(
                            self.get_input_ports()[task_name].get(
                                posixpath.join(self.name, task_name)), name=task_name))
        # Terminate step
        await self.terminate(status)


class ConditionalStep(BaseStep):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__(name, workflow)

    @abstractmethod
    async def _eval(self, inputs: MutableMapping[str, Token]):
        ...

    @abstractmethod
    async def _on_true(self, inputs: MutableMapping[str, Token]):
        ...

    @abstractmethod
    async def _on_false(self, inputs: MutableMapping[str, Token]):
        ...

    async def run(self):
        try:
            if self.input_ports:
                inputs_map = {}
                while True:
                    # Retrieve input tokens
                    inputs = await self._get_inputs(self.get_input_ports())
                    # Check for termination
                    if utils.check_termination(inputs.values()):
                        break
                    # Group inputs by tag
                    _group_by_tag(inputs, inputs_map)
                    # Process tags
                    for tag in list(inputs_map.keys()):
                        if len(inputs_map[tag]) == len(self.input_ports):
                            inputs = inputs_map.pop(tag)
                            # If condition is satisfied (or null)
                            if await self._eval(inputs):
                                await self._on_true(inputs)
                            # Otherwise
                            else:
                                await self._on_false(inputs)
            else:
                # If condition is satisfied (or null)
                if await self._eval({}):
                    await self._on_true({})
                # Otherwise
                else:
                    await self._on_false({})
            await self.terminate(Status.COMPLETED)
        # When receiving a KeyboardInterrupt, propagate it (to allow debugging)
        except KeyboardInterrupt:
            raise
        # When receiving a CancelledError, mark the step as Cancelled
        except CancelledError:
            await self.terminate(Status.CANCELLED)
        except BaseException as e:
            logger.exception(e)
            await self.terminate(Status.FAILED)


class DefaultCommandOutputProcessor(CommandOutputProcessor):

    async def process(self,
                      job: Job,
                      command_output: CommandOutput,
                      connector: Optional[Connector] = None) -> Optional[Token]:
        return Token(
            tag=utils.get_tag(job.inputs.values()),
            value=command_output.value)


class DeployStep(BaseStep):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 deployment_config: DeploymentConfig,
                 connector_port: Optional[ConnectorPort] = None):
        super().__init__(name, workflow)
        self.deployment_config: DeploymentConfig = deployment_config
        self.add_output_port(deployment_config.name, connector_port or workflow.create_port(cls=ConnectorPort))

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> DeployStep:
        params = json.loads(row['params'])
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            deployment_config=await loading_context.load_deployment(context, params['deployment_config']),
            connector_port=cast(ConnectorPort, await loading_context.load_port(context, params['connector_port'])))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        await self.deployment_config.save(context)
        return {**await super()._save_additional_params(context),
                **{'deployment_config': self.deployment_config.persistent_id,
                   'connector_port': self.get_output_port(self.deployment_config.name).persistent_id}}

    def add_output_port(self, name: str, port: ConnectorPort) -> None:
        if not self.output_ports or port.name in self.output_ports:
            super().add_output_port(name, port)
        else:
            raise WorkflowDefinitionException("Deploy step must contain a single output port.")

    def get_output_port(self, name: Optional[str] = None) -> ConnectorPort:
        return cast(ConnectorPort, super().get_output_port(name))

    async def run(self):
        if len(self.output_ports) != 1:
            raise WorkflowDefinitionException("Scatter step must contain a single output port.")
        try:
            if self.input_ports:
                inputs_map = {}
                while True:
                    # Wait for input tokens to be available
                    inputs = await self._get_inputs(self.get_input_ports())
                    # Check for termination
                    if utils.check_termination(inputs.values()):
                        break
                    # Group inputs by tag
                    _group_by_tag(inputs, inputs_map)
                    # Process tags
                    for tag in list(inputs_map.keys()):
                        if len(inputs_map[tag]) == len(self.input_ports):
                            inputs_map.pop(tag)
                            # Deploy the target
                            await self.workflow.context.deployment_manager.deploy(self.deployment_config)
                            # Propagate the connector in the output port
                            self.get_output_port().put(await self._persist_token(
                                token=Token(value=self.deployment_config.name),
                                port=self.get_output_port(),
                                inputs=inputs.values()))
            else:
                # Deploy the target
                await self.workflow.context.deployment_manager.deploy(self.deployment_config)
                # Propagate the connector in the output port
                self.get_output_port().put(await self._persist_token(
                    token=Token(value=self.deployment_config.name),
                    port=self.get_output_port(),
                    inputs=[]))
            await self.terminate(Status.COMPLETED)
        # When receiving a KeyboardInterrupt, propagate it (to allow debugging)
        except KeyboardInterrupt:
            raise
        # When receiving a CancelledError, mark the step as Cancelled
        except CancelledError:
            await self.terminate(Status.CANCELLED)
        except BaseException as e:
            logger.exception(e)
            await self.terminate(Status.FAILED)


class ExecuteStep(BaseStep):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 job_port: JobPort):
        super().__init__(name, workflow)
        self._log_level: int = logging.INFO
        self.command: Optional[Command] = None
        self.output_connectors: MutableMapping[str, str] = {}
        self.output_processors: MutableMapping[str, CommandOutputProcessor] = {}
        self.add_input_port("__job__", job_port)

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> ExecuteStep:
        params = json.loads(row['params'])
        step = cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            job_port=cast(JobPort, await loading_context.load_port(context, params['job_port'])))
        step.output_connectors = params['output_connectors']
        step.output_processors = {k: v for k, v in zip(
            params['output_processors'].keys(),
            await asyncio.gather(*(asyncio.create_task(CommandOutputProcessor.load(context, p, loading_context))
                                   for p in params['output_processors'].values())))}
        return step

    async def _retrieve_output(self,
                               job: Job,
                               output_name: str,
                               output_port: Port,
                               command_output: CommandOutput,
                               connector: Optional[Connector] = None) -> None:
        if (token := await self.output_processors[output_name].process(job, command_output, connector)) is not None:
            output_port.put(await self._persist_token(
                token=token,
                port=output_port,
                inputs=job.inputs.values()))

    async def _run_job(self,
                       inputs: MutableMapping[str, Token],
                       connectors: MutableMapping[str, Connector]) -> Status:
        # Update job
        job = await cast(JobPort, self.get_input_port("__job__")).get_job(self.name)
        if job is None:
            raise WorkflowExecutionException("Step {} received a null job".format(self.name))
        job = Job(
            name=job.name,
            inputs=inputs,
            input_directory=job.input_directory,
            output_directory=job.output_directory,
            tmp_directory=job.tmp_directory)
        logger.debug("Job {name} started".format(name=job.name))
        # Initialise command output with defualt values
        command_output = CommandOutput(value=None, status=Status.FAILED)
        # TODO: Trigger location deployment in case of lazy environments
        try:
            # Execute job
            if not self.terminated:
                self.status = Status.RUNNING
            await self.workflow.context.scheduler.notify_status(job.name, Status.RUNNING)
            command_output = await self.command.execute(job)
            if command_output.status == Status.FAILED:
                logger.error("Job {name} failed {error}".format(
                    name=job.name,
                    error="with error:\n\t{error}".format(error=command_output.value)))
                command_output = await self.workflow.context.failure_manager.handle_failure(job, self, command_output)
        # When receiving a KeyboardInterrupt, propagate it (to allow debugging)
        except KeyboardInterrupt:
            raise
        # When receiving a CancelledError, mark the step as Cancelled
        except CancelledError:
            command_output.status = Status.CANCELLED
            await self.terminate(command_output.status)
        # When receiving a FailureHandling exception, mark the step as Failed
        except FailureHandlingException:
            command_output.status = Status.FAILED
            await self.terminate(command_output.status)
        # When receiving a generic exception, try to handle it
        except BaseException as e:
            logger.exception(e)
            try:
                command_output = await self.workflow.context.failure_manager.handle_exception(job, self, e)
            # If failure cannot be recovered, simply fail
            except BaseException as ie:
                if ie != e:
                    logger.exception(ie)
                command_output.status = Status.FAILED
                await self.terminate(command_output.status)
        finally:
            # Notify completion to scheduler
            await self.workflow.context.scheduler.notify_status(job.name, command_output.status)
        # Retrieve output tokens
        if not self.terminated:
            try:
                await asyncio.gather(*(asyncio.create_task(
                    self._retrieve_output(
                        job=job,
                        output_name=output_name,
                        output_port=self.workflow.ports[output_port],
                        command_output=command_output,
                        connector=connectors.get(output_name))
                ) for output_name, output_port in self.output_ports.items()))
            except BaseException as e:
                logger.exception(e)
                command_output.status = Status.FAILED
        # Return job status
        logger.debug("Job {name} terminated with status {status}".format(
            name=job.name, status=command_output.status.name))
        return command_output.status

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'job_port': self.get_input_port('__job__').persistent_id,
                   'output_connectors': self.output_connectors,
                   'output_processors': {k: v for k, v in zip(
                       self.output_processors.keys(),
                       await asyncio.gather(*(asyncio.create_task(p.save(context))
                                              for p in self.output_processors.values())))}}}

    def add_output_port(self,
                        name: str,
                        port: Port,
                        output_processor: CommandOutputProcessor = None) -> None:
        super().add_output_port(name, port)
        self.output_processors[name] = output_processor or DefaultCommandOutputProcessor(name, self.workflow)

    async def run(self) -> None:
        jobs = []
        # If there are input connector ports, retrieve connectors
        connector_ports = {k: self.get_input_port(v) for k, v in self.output_connectors.items()
                           if isinstance(self.get_input_port(v), ConnectorPort)}
        connectors = {k: v for k, v in zip(connector_ports.keys(), await asyncio.gather(*(
            asyncio.create_task(p.get_connector(posixpath.join(self.name, port_name)))
            for port_name, p in connector_ports.items())))}
        # If there are input ports create jobs until termination token are received
        input_ports = {k: v for k, v in self.get_input_ports().items()
                       if k != "__job__" and not isinstance(v, ConnectorPort)}
        if input_ports:
            inputs_map = {}
            while True:
                # Retrieve input tokens
                inputs = await self._get_inputs(input_ports)
                # Check for termination
                if utils.check_termination(inputs.values()):
                    break
                # Group inputs by tag
                _group_by_tag(inputs, inputs_map)
                # Process tags
                for tag in list(inputs_map.keys()):
                    if len(inputs_map[tag]) == len(input_ports):
                        inputs = inputs_map.pop(tag)
                        # Set status to fireable
                        await self._set_status(Status.FIREABLE)
                        # Run job
                        jobs.append(asyncio.create_task(
                            self._run_job(inputs, connectors),
                            name=utils.random_name()))
        # Otherwise simply run job
        else:
            jobs.append(asyncio.create_task(
                self._run_job({}, connectors),
                name=utils.random_name()))
        # Wait for jobs termination
        statuses = cast(MutableSequence[Status], await asyncio.gather(*jobs))
        # If there are connector ports, retrieve termination tokens from them
        await asyncio.gather(*(
            asyncio.create_task(p.get(posixpath.join(self.name, port_name)))
            for port_name, p in connector_ports.items()))
        # Terminate step
        await self.terminate(_get_step_status(statuses))


class GatherStep(BaseStep):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 depth: int = 1):
        super().__init__(name, workflow)
        self.depth: int = depth
        self.token_map: MutableMapping[str, MutableSequence[Token]] = {}

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> GatherStep:
        params = json.loads(row['params'])
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            depth=params['depth'])

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'depth': self.depth}}

    def add_input_port(self, name: str, port: Port) -> None:
        if not self.input_ports or name in self.input_ports:
            super().add_input_port(name, port)
        else:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))

    def add_output_port(self, name: str, port: Port) -> None:
        if not self.output_ports or port.name in self.output_ports:
            super().add_output_port(name, port)
        else:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))

    async def run(self):
        if len(self.input_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))
        if len(self.output_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))
        input_port = self.get_input_port()
        while True:
            token = await input_port.get(posixpath.join(self.name, next(iter(self.input_ports))))
            if utils.check_termination(token):
                output_port = self.get_output_port()
                for tag, tokens in self.token_map.items():
                    output_port.put(await self._persist_token(
                        token=ListToken(
                            tag=tag,
                            value=sorted(tokens, key=lambda cur: cur.tag)),
                        port=output_port,
                        inputs=tokens))
                break
            else:
                key = '.'.join(token.tag.split('.')[:-self.depth])
                if key not in self.token_map:
                    self.token_map[key] = []
                self.token_map[key].append(token)
        # Terminate step
        await self.terminate(Status.SKIPPED if self.get_output_port().empty() else Status.COMPLETED)


class InputInjectorStep(BaseStep, ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 job_port: JobPort):
        super().__init__(name, workflow)
        self.add_input_port("__job__", job_port)

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> InputInjectorStep:
        params = json.loads(row['params'])
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            job_port=cast(JobPort, await loading_context.load_port(context, params['job_port'])))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'job_port': self.get_input_port('__job__').persistent_id}}

    def add_output_port(self, name: str, port: Port) -> None:
        if not self.output_ports or port.name in self.output_ports:
            super().add_output_port(name, port)
        else:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))

    @abstractmethod
    async def process_input(self,
                            job: Job,
                            token_value: Any) -> Token:
        ...

    async def run(self):
        input_ports = {k: v for k, v in self.get_input_ports().items() if k != "__job__"}
        if len(input_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))
        if len(self.output_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))
        if input_ports:
            while True:
                # Retrieve input token
                token = next(iter((await self._get_inputs(input_ports)).values()))
                # Check for termination
                if utils.check_termination(token):
                    break
                # Retrieve job
                job = await cast(JobPort, self.get_input_port("__job__")).get_job(self.name)
                if job is None:
                    raise WorkflowExecutionException("Step {} received a null job".format(self.name))
                try:
                    await self.workflow.context.scheduler.notify_status(job.name, Status.RUNNING)
                    # Process value and inject token in the output port
                    self.get_output_port().put(await self._persist_token(
                        token=await self.process_input(job, token.value),
                        port=self.get_output_port(),
                        inputs=[token]))
                finally:
                    # Notify completion to scheduler
                    await self.workflow.context.scheduler.notify_status(job.name, Status.COMPLETED)
        # Terminate step
        await self.terminate(Status.SKIPPED if self.get_output_port().empty() else Status.COMPLETED)


class LoopCombinatorStep(CombinatorStep):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 combinator: Combinator):
        super().__init__(name, workflow, combinator)
        self.iteration_terminaton_checklist: MutableMapping[str, Set[str]] = {}

    async def run(self):
        # Set default status to SKIPPED
        status = Status.SKIPPED
        if self.input_ports:
            input_tasks, terminated = [], []
            for port_name, port in self.get_input_ports().items():
                self.iteration_terminaton_checklist[port_name] = set()
                input_tasks.append(asyncio.create_task(
                    port.get(posixpath.join(self.name, port_name)), name=port_name))
            while input_tasks:
                # Wait for the next token
                finished, unfinished = await asyncio.wait(input_tasks, return_when=FIRST_COMPLETED)
                input_tasks = list(unfinished)
                for task in finished:
                    task_name = cast(Task, task).get_name()
                    token = task.result()
                    # If a TerminationToken is received, the corresponding port terminated its outputs
                    if utils.check_termination(token):
                        logger.debug("Step {} received termination token for port {}".format(
                            self.name, task_name))
                        terminated.append(task_name)
                    # If an IterationTerminationToken is received, mark the corresponding iteration as terminated
                    elif utils.check_iteration_termination(token):
                        if token.tag in self.iteration_terminaton_checklist[task_name]:
                            logger.debug("Step {} received iteration termination token {} for port {}".format(
                                self.name, token.tag, task_name))
                            self.iteration_terminaton_checklist[task_name].remove(token.tag)
                    # Otherwise, build combination and set default status to COMPLETED
                    else:
                        logger.debug("Step {} received token {} on port {}".format(
                            self.name, token.tag, task_name))
                        status = Status.COMPLETED
                        if '.'.join(token.tag.split('.')[:-1]) not in self.iteration_terminaton_checklist[task_name]:
                            self.iteration_terminaton_checklist[task_name].add(token.tag)
                        async for schema in cast(AsyncIterable, self.combinator.combine(task_name, token)):
                            for port_name, token in schema.items():
                                self.get_output_port(port_name).put(await self._persist_token(
                                    token=token,
                                    port=self.get_output_port(port_name),
                                    inputs=schema.values()))
                    # Create a new task in place of the completed one if the port is not terminated
                    if not (task_name in terminated and len(self.iteration_terminaton_checklist[task_name]) == 0):
                        input_tasks.append(asyncio.create_task(
                            self.get_input_ports()[task_name].get(
                                posixpath.join(self.name, task_name)), name=task_name))
        # Terminate step
        await self.terminate(status)


class LoopOutputStep(BaseStep, ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__(name, workflow)
        self.token_map: MutableMapping[str, MutableSequence[Token]] = {}
        self.size_map: MutableMapping[str, int] = {}
        self.termination_map: MutableMapping[str, bool] = {}

    @abstractmethod
    async def _process_output(self, tag: str) -> Token:
        ...

    def add_input_port(self, name: str, port: Port) -> None:
        if not self.input_ports or name in self.input_ports:
            super().add_input_port(name, port)
        else:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))

    def add_output_port(self, name: str, port: Port) -> None:
        if not self.output_ports or port.name in self.output_ports:
            super().add_output_port(name, port)
        else:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))

    async def run(self):
        if len(self.input_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single input port.".format(self.name))
        if len(self.output_ports) != 1:
            raise WorkflowDefinitionException("{} step must contain a single output port.".format(self.name))
        input_port = self.get_input_port()
        while True:
            token = await input_port.get(posixpath.join(self.name, next(iter(self.input_ports))))
            prefix = '.'.join(token.tag.split('.')[:-1])
            # If a TerminationToken is received, terminate the step
            if utils.check_termination(token):
                logger.debug("Step {} received termination token".format(self.name))
                self.termination_map = {k: len(self.token_map[k]) == self.size_map.get(k, -1) for k in self.token_map}
            # If an IterationTerminationToken is received, process loop output for the current port
            elif utils.check_iteration_termination(token):
                logger.debug("Step {} received iteration termination token {}.".format(
                    self.name, token.tag))
                self.size_map[prefix] = int(token.tag.split('.')[-1])
            # Otherwise, store the new token in the map
            else:
                logger.debug("Step {} received token {}.".format(
                    self.name, token.tag))
                if prefix not in self.token_map:
                    self.token_map[prefix] = []
                self.token_map[prefix].append(token)
            if len(self.token_map.get(prefix, [])) == self.size_map.get(prefix, -1):
                self.get_output_port().put(await self._persist_token(
                    token=await self._process_output(prefix),
                    port=self.get_output_port(),
                    inputs=self.token_map.get(prefix)))
            # If all iterations are terminated, terminate the step
            if self.termination_map and all(self.termination_map):
                break
        # Terminate step
        await self.terminate(Status.SKIPPED if self.get_output_port().empty() else Status.COMPLETED)


class ScheduleStep(BaseStep):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 target: Target,
                 connector_port: ConnectorPort,
                 job_port: Optional[JobPort] = None,
                 hardware_requirement: Optional[HardwareRequirement] = None,
                 input_directory: Optional[str] = None,
                 output_directory: Optional[str] = None,
                 tmp_directory: Optional[str] = None):
        super().__init__(name, workflow)
        self.target: Target = target
        self.hardware_requirement: Optional[HardwareRequirement] = hardware_requirement
        self.input_directory: Optional[str] = input_directory
        self.output_directory: Optional[str] = output_directory
        self.tmp_directory: Optional[str] = tmp_directory
        self.add_input_port("__connector__", connector_port)
        self.add_output_port("__job__", job_port or workflow.create_port(cls=JobPort))

    def _get_directory(self,
                       path_processor,
                       directory: Optional[str]):
        return directory or path_processor.join(self.target.workdir, utils.random_name())

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> ScheduleStep:
        params = json.loads(row['params'])
        if hardware_requirement := params.get('hardware_requirement'):
            hardware_requirement = await HardwareRequirement.load(
                context, hardware_requirement, loading_context)
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            target=await loading_context.load_target(context, params['target']),
            connector_port=cast(ConnectorPort, await loading_context.load_port(context, params['connector_port'])),
            job_port=cast(JobPort, await loading_context.load_port(context, params['job_port'])),
            hardware_requirement=hardware_requirement,
            input_directory=params['input_directory'],
            output_directory=params['output_directory'],
            tmp_directory=params['tmp_directory'])

    async def _propagate_job(self,
                             connector: Connector,
                             locations: MutableSequence[str],
                             job: Job):
        # Create directories
        await remotepath.mkdirs(
            connector=connector,
            locations=locations,
            paths=[job.input_directory, job.output_directory, job.tmp_directory])
        # Propagate job
        self.get_output_port().put(await self._persist_token(
            token=JobToken(value=job),
            port=self.get_output_port(),
            inputs=job.inputs.values()))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        await self.target.save(context)
        params = {**await super()._save_additional_params(context),
                  **{'connector_port': self.get_input_port('__connector__').persistent_id,
                     'job_port': self.get_output_port('__job__').persistent_id,
                     'target': self.target.persistent_id,
                     'input_directory': self.input_directory,
                     'output_directory': self.output_directory,
                     'tmp_directory': self.tmp_directory}}
        if self.hardware_requirement:
            params['hardware_requirement'] = await self.hardware_requirement.save(context)
        return params

    def get_output_port(self, name: Optional[str] = None) -> JobPort:
        return cast(JobPort, super().get_output_port(name))

    async def run(self):
        try:
            # Retrieve connector
            connector = await cast(ConnectorPort, self.get_input_port("__connector__")).get_connector(self.name)
            path_processor = utils.get_path_processor(connector)
            # If there are input ports
            input_ports = {k: v for k, v in self.get_input_ports().items() if k != "__connector__"}
            if input_ports:
                inputs_map = {}
                while True:
                    # Retrieve input tokens
                    inputs = await self._get_inputs(input_ports)
                    # Check for termination
                    if utils.check_termination(inputs.values()):
                        break
                    # Group inputs by tag
                    _group_by_tag(inputs, inputs_map)
                    # Process tags
                    for tag in list(inputs_map.keys()):
                        if len(inputs_map[tag]) == len(input_ports):
                            inputs = inputs_map.pop(tag)
                            # Create Job
                            job = Job(
                                name=random_name(),
                                inputs=inputs,
                                input_directory=self._get_directory(path_processor, self.input_directory),
                                output_directory=self._get_directory(path_processor, self.output_directory),
                                tmp_directory=self._get_directory(path_processor, self.tmp_directory))
                            # Schedule
                            hardware_requirement = (self.hardware_requirement.eval(inputs)
                                                    if self.hardware_requirement else None)
                            await self.workflow.context.scheduler.schedule(job, self.target, hardware_requirement)
                            locations = self.workflow.context.scheduler.get_locations(job.name)
                            await self._propagate_job(connector, locations, job)
            else:
                # Create Job
                job = Job(
                    name=random_name(),
                    inputs={},
                    input_directory=self._get_directory(path_processor, self.input_directory),
                    output_directory=self._get_directory(path_processor, self.output_directory),
                    tmp_directory=self._get_directory(path_processor, self.tmp_directory))
                # Schedule
                await self.workflow.context.scheduler.schedule(
                    job, self.target, self.hardware_requirement.eval({}) if self.hardware_requirement else None)
                locations = self.workflow.context.scheduler.get_locations(job.name)
                await self._propagate_job(connector, locations, job)
            await self.terminate(Status.SKIPPED if self.get_output_port().empty() else Status.COMPLETED)
        # When receiving a KeyboardInterrupt, propagate it (to allow debugging)
        except KeyboardInterrupt:
            raise
        # When receiving a CancelledError, mark the step as Cancelled
        except CancelledError:
            await self.terminate(Status.CANCELLED)
        except BaseException as e:
            logger.exception(e)
            await self.terminate(Status.FAILED)


class ScatterStep(BaseStep):

    async def _scatter(self, token: Token):
        if isinstance(token.value, Token):
            await self._scatter(token.value)
        elif isinstance(token, ListToken):
            output_port = self.get_output_port()
            for i, t in enumerate(token.value):
                output_port.put(await self._persist_token(
                    token=t.retag(token.tag + '.' + str(i)),
                    port=output_port,
                    inputs=[token]))
        else:
            raise WorkflowDefinitionException("Scatter ports require iterable inputs")

    def add_input_port(self, name: str, port: Port) -> None:
        if not self.input_ports:
            super().add_input_port(name, port)
        else:
            raise WorkflowDefinitionException("Scatter step must contain a single input port.")

    def add_output_port(self, name: str, port: Port) -> None:
        if not self.output_ports or port.name in self.output_ports:
            super().add_output_port(name, port)
        else:
            raise WorkflowDefinitionException("Scatter step must contain a single output port.")

    async def run(self):
        if len(self.input_ports) != 1:
            raise WorkflowDefinitionException("Scatter step must contain a single input port.")
        if len(self.output_ports) != 1:
            raise WorkflowDefinitionException("Scatter step must contain a single output port.")
        input_port = self.get_input_port()
        output_port = self.get_output_port()
        while True:
            token = await input_port.get(posixpath.join(self.name, next(iter(self.input_ports))))
            if isinstance(token, TerminationToken):
                break
            else:
                await self._scatter(token)
        # Terminate step
        await self.terminate(Status.SKIPPED if output_port.empty() else Status.COMPLETED)


class TransferStep(BaseStep, ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow,
                 job_port: JobPort):
        super().__init__(name, workflow)
        self.add_input_port("__job__", job_port)

    @classmethod
    async def _load(cls,
                    context: StreamFlowContext,
                    row: MutableMapping[str, Any],
                    loading_context: DatabaseLoadingContext) -> TransferStep:
        params = json.loads(row['params'])
        return cls(
            name=row['name'],
            workflow=await loading_context.load_workflow(context, row['workflow']),
            job_port=cast(JobPort, await loading_context.load_port(context, params['job_port'])))

    async def _save_additional_params(self, context: StreamFlowContext) -> MutableMapping[str, Any]:
        return {**await super()._save_additional_params(context),
                **{'job_port': self.get_input_port('__job__').persistent_id}}

    async def run(self):
        # Set default status as SKIPPED
        status = Status.SKIPPED
        # Retrieve input ports
        input_ports = {k: v for k, v in self.get_input_ports().items() if k != "__job__"}
        if input_ports:
            inputs_map = {}
            try:
                while True:
                    # Retrieve input tokens
                    inputs = await self._get_inputs(input_ports)
                    # Check for termination
                    if utils.check_termination(inputs.values()):
                        break
                    # Group inputs by tag
                    _group_by_tag(inputs, inputs_map)
                    # Process tags
                    for tag in list(inputs_map.keys()):
                        if len(inputs_map[tag]) == len(input_ports):
                            inputs = inputs_map.pop(tag)
                            # Retrieve job
                            job = await cast(JobPort, self.get_input_port('__job__')).get_job(self.name)
                            if job is None:
                                raise WorkflowExecutionException("Step {} received a null job".format(self.name))
                            # Change default status to COMPLETED
                            status = Status.COMPLETED
                            # Transfer token
                            for port_name, token in inputs.items():
                                self.get_output_port(port_name).put(await self._persist_token(
                                    token=await self.transfer(job, token),
                                    port=self.get_output_port(port_name),
                                    inputs=inputs.values()))
            # When receiving a KeyboardInterrupt, propagate it (to allow debugging)
            except KeyboardInterrupt:
                raise
            # When receiving a CancelledError, mark the step as Cancelled
            except CancelledError:
                await self.terminate(Status.CANCELLED)
            except BaseException as e:
                logger.exception(e)
                await self.terminate(Status.FAILED)
        # Terminate step
        await self.terminate(status)

    @abstractmethod
    async def transfer(self, job: Job, token: Token) -> Token:
        ...


class Transformer(BaseStep, ABC):

    def __init__(self,
                 name: str,
                 workflow: Workflow):
        super().__init__(name, workflow)

    async def run(self):
        try:
            if self.input_ports:
                inputs_map = {}
                while True:
                    # Retrieve input tokens
                    inputs = await self._get_inputs(self.get_input_ports())
                    # Check for termination
                    if utils.check_termination(inputs.values()):
                        break
                    # Group inputs by tag
                    _group_by_tag(inputs, inputs_map)
                    # Process tags
                    for tag in list(inputs_map.keys()):
                        if len(inputs_map[tag]) == len(self.input_ports):
                            inputs = inputs_map.pop(tag)
                            # Check for iteration termination and propagate
                            if utils.check_iteration_termination(inputs.values()):
                                for port_name, token in inputs.items():
                                    self.get_output_port(port_name).put(await self._persist_token(
                                        token=token,
                                        port=self.get_output_port(port_name),
                                        inputs=inputs.values()))
                            # Otherwise, apply transformation and propagate outputs
                            else:
                                for port_name, token in (await self.transform(inputs)).items():
                                    self.get_output_port(port_name).put(await self._persist_token(
                                        token=token,
                                        port=self.get_output_port(port_name),
                                        inputs=inputs.values()))
            else:
                for port_name, token in (await self.transform({})).items():
                    self.get_output_port(port_name).put(await self._persist_token(
                        token=token,
                        port=self.get_output_port(port_name),
                        inputs=[]))
            # Terminate step
            await self.terminate(Status.SKIPPED if any(p.empty() for p in self.get_output_ports().values())
                                 else Status.COMPLETED)
        # When receiving a KeyboardInterrupt, propagate it (to allow debugging)
        except KeyboardInterrupt:
            raise
        # When receiving a CancelledError, mark the step as Cancelled
        except CancelledError:
            await self.terminate(Status.CANCELLED)
        except BaseException as e:
            logger.exception(e)
            await self.terminate(Status.FAILED)

    @abstractmethod
    async def transform(self, inputs: MutableMapping[str, Token]) -> MutableMapping[str, Token]:
        ...
